import random
import time
import uuid
from pprint import pprint
from typing import Any, List, Optional

import maya
from addict import Dict
from loguru import logger

from jamboree import Jamboree
from jamboree.handlers.abstracted.search.meta import MetadataSearchHandler
from jamboree.handlers.complex.backtestable import BacktestBlobHandler
from jamboree.handlers.complex.meta import MetaHandler
from jamboree.handlers.complex.metric import MetricHandler
from jamboree.middleware.procedures import (
    ProcedureAbstract,
    ProcedureManagement,
)
from jamboree.utils.support.search import querying

logger.disable('jamboree')


class FileEngine(BacktestBlobHandler):
    """ 
        # FileEngine

        Mixing:
            * Native Backtesting
            * BlobHandler
            * Metrics
            * ContextManager
            * Search

        The goal of this handler is to be able to have something generic you can use for all of our backtesting needs with blobs.

        The idea behind it is that we'd need to have raw blobs to manage things like the following:
            1. Saving and loading strategies
            2. Saving and loading machine learning models
        
        For those two we'll need to do the following:

            1. Search through models
            2. Store metrics for a given file
                * So we can see the performance for a given item over time (such as a machine learning calculation/etc)
            3. Get specific files by ID
            4. Set and get settings of objects
            5. Make sure we don't miss anything
                * Saving metrics for a strategy or model
                * Saving our models and stratgies after using them. 
                    * This is especially important if we haven't seen certain data before.
                * Running other clean up commands at the end.
        
        The core solution for this is to create an integrated setup that we can inherit from and run in a common way.
        
        The second is to use a context manager to load information and use it immediately. 
        
        If we have an online learning algorithm we'd have a `ModelEngine`, which would inherit from the `FileEngine` and open the context manager. 

        We'd then do the work we could do the following command:

            ```
                with compute_engine as model:
                    model.partial_fit(x, y)
            ```
    
    """

    def __init__(self, processor=None, **kwargs) -> None:
        super(FileEngine, self).__init__()
        self.entity = "file_engine"  # Gen model represents general model
        self.required = {
            "name": str,
            "category": str,
            "subcategories": dict,  # other information regarding the library type.
            # Search specific information
            "metatype": str,
            "submetatype": str,
            "abbreviation": str,
        }
        self["metatype"] = self.entity

        if processor is not None:
            self.processor = processor
        self.metaid = ""
        self.initialize(**kwargs)

        self.__name:str = ""
        self.__category:str = ""
        self.__subcategories:dict = {}
        self.__metatype:str = ""
        self.__submetatype:str = ""
        self.__abbreviation:str = ""
    
    @property
    def name(self) -> str:
        """The name property."""
        return self["name"]
    
    @name.setter
    def name(self, value:str):
        self["name"] = value

    @property
    def category(self) -> str:
        """The category property."""
        return self["category"]


    @category.setter
    def category(self, value:str):
        self["category"] = value


    @property
    def subcategories(self) -> dict:
        """The subcategories property."""
        return self["subcategories"]


    @subcategories.setter
    def subcategories(self, value:dict):
        self["subcategories"] = value
    

    @property
    def metatype(self) -> str:
        """The metatype property."""
        return self.entity
    

    @metatype.setter
    def metatype(self, value:str):
        self.entity = value
        self["metatype"] = value
    
    @property
    def submetatype(self) -> str:
        """The submetatype property."""
        return self['submetatype']
    
    @submetatype.setter
    def submetatype(self, value:str):
        self['submetatype'] = value
    

    @property
    def abbreviation(self)->str:
        """The abbreviation property."""
        return self["abbreviation"]

    @abbreviation.setter
    def abbreviation(self, value:str):
        self["abbreviation"] = value

    """ 
        Context Manager

        with compute_engine as model:
            prediction = self.model.fit_predict(data) # we'd have preprocessing strategies
    """

    def __enter__(self):
        self.open_context()
        return self.enterable()

    def __exit__(self, exc_type, exc_val, exc_tb):
        # make sure the dbconnection gets closed
        self.close_context()

    def enterable(self):
        """ Return the object we want to enter into """
        raise NotImplementedError("Enterable not added yet.")

    def open_context(self):
        """ Take the model from the model procedure and save it. """
        raise NotImplementedError("Open context hasn't had anything added")

    def close_context(self):
        """ Take the model from the model procedure and save it. """
        raise NotImplementedError("Close context hasn't had anything added.")

    """
        Initialize everything for the user
    """

    def initialize(self, **kwargs):
        self.init_handlers()
        self.init_settings(**kwargs)
        self.init_required(**kwargs)
        self.init_specialized(**kwargs)

    def init_handlers(self):
        self._metrics: MetricHandler = MetricHandler()
        self._metasearch: MetadataSearchHandler = MetadataSearchHandler()

    def init_settings(self, **kwargs):
        """ All of the general settings that we need to initialize """
        # self.metadata.settings
        self.online = kwargs.get("online", False)
        self.bfile = kwargs.get("blobfile", None)
        self.file_reset = False
        self.searchable_items = kwargs.get(
            "searchable", ["name", "subcategories", "abbreviation"]
        )
        self.pm: Optional[ProcedureManagement] = kwargs.get("proc_management", None)

    def init_required(self, **kwargs):
        if len(kwargs) == 0:
            return
        for k, v in self.required.items():
            if k in kwargs:
                value = kwargs.get(k)
                if isinstance(value, v):
                    self[k] = value

    def init_specialized(self, **kwargs):
        """ Initialize all of the highly specific parts """
        self.current_procedure = None

    """ Properties """

    """ 
        # Handler Properties 
        

        Other handlers to interact with other parts of the system.

        1. `self.metrics` to interact with the metrics (logging)
        2. `self.metadata` to interact with the FileEngine's settings and capacity to find information
        3. `self.search` to handle search specified to the key handler. 

        Each handler has its own information attached.
    """

    @property
    def metrics(self):
        self._metrics.processor = self.processor
        self._metrics["name"] = self["name"]
        self._metrics["category"] = self["category"]
        self._metrics["subcategories"] = self["subcategories"]
        self._metrics.episode = self.episode
        self._metrics.live = self.live
        self._metrics.time = self.time
        self._metrics.reset()
        return self._metrics

    @property
    def metadata(self):
        self._meta.processor        = self.processor
        self._meta["name"]          = self["name"]
        self._meta["category"]      = self["category"]
        self._meta["subcategories"] = self["subcategories"]
        self._meta["metatype"]      = self.entity
        self._meta["submetatype"]   = self['submetatype']
        self._meta["abbreviation"]  = self["abbreviation"]
        return self._meta

    @property
    def search(self) -> MetadataSearchHandler:
        self._metasearch.reset()
        self._metasearch["category"] = querying.text.exact(self["category"])
        self._metasearch["metatype"] = querying.text.exact(self.entity)
        self._metasearch["submetatype"] = querying.text.exact(self["submetatype"])
        self._metasearch.processor = self.processor
        return self._metasearch

    """ 
        # Key Variables 
        ---

        The core variables to handle the functionality of the file engine. 

        1. Checking to see if a file is in storage
        2. Checking to see if we've modified a variable since interacting with storage
        3. Setting verification methods and procedure management
    """

    @property
    def is_exist(self) -> bool:
        """ Checks to see if a model exist in storage."""
        existence = self.absolute_exists()
        return existence

    @property
    def is_exist_forced(self) -> bool:
        """ Checks to see if a model exist in storage."""
        self.changed_since_command = True
        existence = self.absolute_exists()
        return existence

    @property
    def changed(self) -> bool:
        return self.changed_since_command
    @property
    def procedure(self) -> Optional["ProcedureAbstract"]:
        raise NotImplementedError("Procedure has not been implemented")

    @procedure.setter
    def procedure(self, _procedure: "ProcedureAbstract"):
        """ Set the access procedure"""
        self.current_procedure = _procedure

    @property
    def is_procedure(self):
        return self.current_procedure is not None

    @property
    def is_proc(self) -> bool:
        """ Checks to see if procedure has been set"""
        return not self.current_procedure

    @property
    def is_local_file(self) -> bool:
        """ Boolean explaining if we've set a file locally. Files can be set either through pulling from the DB or manually"""
        # if the blob file has been set locally, set the response to true
        return self.blobfile is not None
    @property
    def blobfile(self):
        return self.bfile

    @blobfile.setter
    def blobfile(self, _file):
        self.bfile = _file

    """
        # Database Commands

        * Save file
        * Load file
    """

    def verify(self):
        self.pm.isattr(self)


    def load_file(self):
        loaded_file = self.last()
        return loaded_file

    def save_file(self, _file: Any, alt={}):
        self.save(_file, alt=alt, is_overwrite=self.online)


    """
        # SEARCH

        Search related functions from metadata.

        * load_search 
            * Get a MetadataSearchHandler that automatically adds the searchable items into the data.
        
        * search_all
            * A generalized search for all of our datasets given a general term and 
    """

    def loaded_search(self, **kwargs) -> MetadataSearchHandler:
        """ 
            # LOAD SEARCH

            * Get a meatadatasearchandler that automatically adds the searchable fields into the search field.
            * Ensure to change self.searchable_items to change what we're allowed to search through
        """
        self.clear()
        current_search = self.search
        for field in self.searchable_items:
            if field in kwargs:
                current_value = kwargs.get(field)
                if current_value is not None:
                    current_search[field] = current_value
        return current_search

    def search_all(self, general: str, **kwargs):
        """ 
            # SEARCH ALL

            * Search through all of the records. 
            * To be used to find through files from the UI.
        """
        current_search = self.loaded_search(**kwargs)
        current_search.general = general
        return current_search.find()

    def file_from_dict(self, item: Dict):
        reloaded = FileEngine(
            processor=self.processor,
            name=item.name,
            category=item.category,
            subcategories=item.subcategories,
            submetatype=item.submetatype,
            abbreviation=item.abbreviation,
        )
        return reloaded
    def first(self, **kwargs) -> "FileEngine":
        """ Find a model by metadata. If there are models, get the one that exist. Otherwise return None"""
        current_search = self.loaded_search(**kwargs)

        items = current_search.find()
        if len(items) == 0:
            raise AttributeError("No model metadata found with those exact parameters")

        first = items[0]
        first = Dict(**first)

        reloaded = self.file_from_dict(first)
        reloaded.reset()

        return reloaded

    def pick(self, _id: str):
        """ Get a specific item from the system by first using the id generated"""
        self.clear()
        current_search = self.search
        item = current_search.pick(_id)
        if item is None:
            raise AttributeError("File Metadata doesn't exist")

        item = Dict(**item)
        reloaded = self.file_from_dict(item)
        reloaded.reset()
        return reloaded

    """ Reset Commands """

    def custom_post_load(self, item):
        """ Customize what happens after we load it """
        raise NotImplementedError("You need to figure out how to manage a loaded file")

    def reset_exist(self, **kwargs):
        """ Reset logic for if a file already exists """
        item = self.load_file()
        self.custom_post_load(item)
        return item

    def reset_noexist(self):
        """
            Save a file if it doesn't exist yet. Add procedure filter. Load function won't check for this.
        """
        if self.is_local_file:
            # logger.success("Starting to save file")
            if isinstance(self.blobfile, ProcedureAbstract):
                extracted = self.blobfile.extract()
                self.save_file(extracted)
                return
            self.save_file(self.blobfile)
            # logger.success("File officially saved")
        else:
            raise AttributeError(
                "You either need to have a procedure loaded or a pre-existing model"
            )

    def reset_file(self):
        """
            # RESET FILE

            Get the files. 
        """
        self.verify()

        if self.file_reset and not self.changed_since_command:
            return
        if self.is_exist_forced:
            self.reset_exist()
        else:
            self.reset_noexist()
        self.file_reset = True
        self.changed_since_command = False

    def reset(self):
        super().reset()
        if self.is_exist_forced:
            self.metaid = self.metadata.reset()
        self.reset_file()
        self.model_reset = False
        self.metrics.reset()


import json
import pprint
import time
import uuid
from copy import deepcopy
from typing import Any, Dict, List

import crayons as cy
import maya
import pandas_datareader.data as web
import ujson
from loguru import logger

from jamboree import JamboreeNew
from jamboree.handlers.abstracted.search import MetadataSearchHandler
from jamboree.handlers.complex.meta import MetaHandler
from jamboree.handlers.default import DataHandler, DBHandler, TimeHandler, Access
from jamboree.middleware.processors import DataProcessorsAbstract, DynamicResample
from jamboree.utils.context import example_space
from jamboree.utils.core import consistent_hash, consistent_unhash
from jamboree.utils.support.search import querying



"""
    NOTE: Will probably inherit this to fix it in private.
    This is dog shit!!!!
    
    A lot of this will be replaced using a task graphs (to manage relationships), redisearch aggregations and pipelines (crafting a tree of interactions)

"""


class MultiDataManagement(Access):
    """ 
        # Multi-Data Handler
        ---

        Here we define how different data sources will be pulled at the same time. 
        
        Given a set of parameters we can select which data sources we want to pull for a given problem. 


        ## Why is this useful?
        ---
        This is useful because we're able to pull together any number of data sources we have stored inside of our database.

        Some usecases of this would be:

        - Getting multiple assets to define a portfolio allocation.
        - Getting different asset classes define pairs to trade on.
        - Getting a superset of different asset classes to predict movements between them.
        - Getting a predefied pair to determine which signals should be used.
    """

    def __init__(self, **kwargs):
        super().__init__()

        self.entity = "multi_data_management"
        """ 
            # Required variables
            ---

            - Set name This is the name of the data sets we're going to have
        """
        self.required = {
            "name": str,
            "category": str,
            "subcategories": dict,
            "metatype": str,
            "submetatype": str,
            "abbreviation": str,
        }
        self.is_robust = False
        self.initialize(**kwargs)
        self.data_handler_list: List[
            DataHandler
        ] = []  # Store the dataset objects we can access at once without redeclaring
        self.dup_check_list = []  # use to check for duplicates in dataset
        
        
        self._meta = MetaHandler()
        self._metasearch = MetadataSearchHandler()
        self.metaid: str = ""
        self.description = kwargs.get("description", None)

    def initialize(self, **kwargs):
        """ A more concentrated place to initialize everything"""
        self.init_handlers()
        self.init_default(**kwargs)
        self.init_required(**kwargs)

    def init_default(self, **kwargs):
        self._episode: str = kwargs.get("episode", uuid.uuid4().hex)
        self._is_live: bool = kwargs.get("live", False)
        self._preprocessor = kwargs.get("preprocessor", DynamicResample("data"))

        self.is_real_filter: bool = kwargs.get("real_filter", True)
        self.metatype = self.entity
        self.category = "universe"
        self.submetatype = "price_bag"

        # self["metatype"] = self.entity
        # self["category"] = "universe"
        # self["submetatype"] = "price_bag"
        
        self.is_event = False

    def init_required(self, **kwargs):
        if len(kwargs) == 0:
            return
        for k, v in self.required.items():
            if k in kwargs:
                value = kwargs.get(k)
                if isinstance(value, v):
                    self[k] = value

    def init_handlers(self):
        self.datasethandler: DataHandler = DataHandler()
        self._time: TimeHandler = TimeHandler()

    @property
    def episode(self) -> str:
        return self._episode

    @episode.setter
    def episode(self, _episode: str):
        self._episode = _episode

    @property
    def live(self) -> bool:
        return self._is_live

    @live.setter
    def live(self, _live: bool):
        self._is_live = _live

    @property
    def sources(self) -> list:
        source_dict = self.latest_dataset_list()
        source_list = source_dict.get("sources", [])
        return source_list

    @property
    def datasets(self) -> List[DataHandler]:
        return self.data_handler_list

    @property
    def time(self) -> "TimeHandler":
        self._time.event = self.event
        self._time.processor = self.processor
        self._time["episode"] = self.episode
        self._time["live"] = self.live
        return self._time

    @property
    def metadata(self):
        self._meta.processor = self.processor
        self._meta["name"] = self.name
        self._meta["category"] = self.category
        self._meta["subcategories"] = self.subcategories
        self._meta["metatype"] = self.entity
        self._meta["submetatype"] = self.submetatype
        self._meta["abbreviation"] = self.abbreviation
        self._meta.description = self.description
        return self._meta

    @property
    def search(self):
        self._metasearch.reset()
        self._metasearch["category"] = querying.text.exact(self.category)
        self._metasearch["metatype"] = querying.text.exact(self.entity)
        self._metasearch["submetatype"] = querying.text.exact(self.submetatype)
        self._metasearch.processor = self.processor
        return self._metasearch

    @property
    def preprocessor(self) -> DataProcessorsAbstract:
        return self._preprocessor

    @preprocessor.setter
    def preprocessor(self, _preprocessor: DataProcessorsAbstract):
        self._preprocessor = _preprocessor

    @property
    def is_next(self) -> bool:
        """ Determine if anything is next in the head"""
        processor = self.processor
        is_live_list: List[bool] = []
        for ds in self.datasets:
            ds.processor = processor
            ds.episode   = self.episode
            ds.live      = self.live
            ds.is_robust = self.is_robust
            is_live_list.append(ds.is_next)

        # Use all to determine if the values are falsey or not
        return all(is_live_list)


    @property
    def source_ids(self) -> List[str]:
        """Get SourceIds

        Get the metadata id for the individual data sources inside of the multi-data set.

        Returns
        -------
        List[str]
            List of hex ids
        """


        _source_ids = []
        for source in self.sources:
            sourceid = self._get_source_id(source)
            if sourceid is not None:
                _source_ids.append(sourceid)
        return _source_ids

    def allvalid(self, item: dict):

        name = item.get("name", None)
        category = item.get("category", None)
        subcategories = item.get("subcategories", None)
        submetatype = item.get("submetatype", None)
        abbreviation = item.get("abbreviation", None)
        if None in [name, category, subcategories, submetatype, abbreviation]:
            return False
        return True

    def add_multiple_data_sources(
        self, sources: List[Dict[str, Any]], alt={}, allow_bypass=False
    ):
        """ Add a dataset list"""
        if allow_bypass == True:
            self.save_dataset_list(sources)
            return
        is_valid = self._validate_added_sources(sources)
        if is_valid:
            self.save_dataset_list(sources)

    def remove_multiple_datasources(
        self, rsources: List[Dict[str, Any]], alt={}, allow_bypass=False
    ):

        existing = self.sources
        if len(existing) == 0 or len(rsources) == 0:
            return

        rhashes = set(consistent_hash(i) for i in rsources)
        ohashes = set(consistent_hash(i) for i in existing)

        remaining_hash = ohashes - rhashes
        remaining = [consistent_unhash(x) for x in remaining_hash]
        sources_dict = {"sources": remaining}
        self.save(sources_dict)

    def add_data_source(self, source: Dict[str, Any]):
        if not isinstance(source, dict):
            logger.error("Not a dictionary. Skipping ... ")
            return
        source_list = [source]
        self.add_multiple_data_sources(source_list)

    def remove_data_source(self, source: Dict[str, Any]):
        if not isinstance(source, dict):
            logger.error("Not a dictionary. Skipping ... ")
            return
        source_list = [source]
        self.remove_multiple_datasources(source_list)

    def add_dataset_handler(self):
        _copy = self.datasethandler.copy()

        str_copy = str(_copy)
        if str_copy in self.dup_check_list:
            return
        self.dup_check_list.append(str_copy)
        self.datasethandler.event = self.event
        self.datasethandler.processor = self.processor
        self.data_handler_list.append(_copy)

    def _is_data_exist(self, source: dict) -> bool:
        """ Check to see if the data exist when putting list of datahandlers together"""
        self.datasethandler.event = self.event
        self.datasethandler.processor = self.processor
        self.datasethandler["name"] = source["name"]
        self.datasethandler["category"] = source["category"]
        self.datasethandler["subcategories"] = source["subcategories"]
        self.datasethandler["submetatype"] = source["submetatype"]
        self.datasethandler["abbreviation"] = source["abbreviation"]
        count = self.datasethandler.count()
        not_zero = count != 0
        if not_zero:
            self.add_dataset_handler()
        return not_zero
    

    def _get_source_id(self, source: dict) -> bool:
        """ Check to see if the data exist when putting list of datahandlers together"""
        self.datasethandler.event               = self.event
        self.datasethandler.processor           = self.processor
        self.datasethandler["name"]             = source["name"]
        self.datasethandler["category"]         = source["category"]
        self.datasethandler["subcategories"]    = source["subcategories"]
        self.datasethandler["submetatype"]      = source["submetatype"]
        self.datasethandler["abbreviation"]     = source["abbreviation"]
        count = self.datasethandler.count()
        not_zero = count != 0

        if not_zero:
            return self.datasethandler.reset()
        return None

    def _add_wo_duplicates(self, original_list: list, new_list: list):
        original_set = set(ujson.dumps(i, sort_keys=True) for i in original_list)
        for item in new_list:
            frozen = ujson.dumps(item, sort_keys=True)
            original_set.add(frozen)
        return [ujson.loads(x) for x in original_set]

    def _remove_invalid_dataset_formats(self, original_list: List[Dict[str, Any]]):
        valid_list = []
        for original in original_list:
            if not self.allvalid(original):
                continue
            valid_list.append(original)
        return valid_list

    def _filter_non_existing_datasets(self, _datasets):
        all_sets = []
        for source in _datasets:
            exist = self._is_data_exist(source)
            if exist == True:
                all_sets.append(source)
        return all_sets

    def _validate_added_sources(self, sources: List[Dict[str, Any]]):
        if not isinstance(sources, list):
            logger.error("Sources is not a list. Skipping ...")
            return False

        if len(sources) == 0:
            logger.error("Sources is empty. Skipping ...")
            return False

        for source in sources:
            if not isinstance(source, dict):
                logger.error(
                    "An item inside of the sources list is not a dictionary. Skipping ..."
                )
                return False

            keys = source.keys()
            if len(keys) == 0:
                logger.error("One of the dictionaries doesn't have a key. Skipping ...")
                return False

            for key in keys:
                if not isinstance(key, str):
                    logger.error(
                        "Dude, stop messing up. One of the keys isn't a string. Skipping ..."
                    )
                    return False
        return True

    # -------------------------------------------------------------------------------------------
    # -------------------------------------------------------------------------------------------
    # --------------------------------- I/O Commands --------------------------------------------
    # -------------------------------------------------------------------------------------------
    # -------------------------------------------------------------------------------------------
    def save_dataset_list(self, sources: List[Dict[str, Any]]):
        self.check()
        # We store the sources list into a dictionary to make it easier for Jamboree to handle
        latest_sources = self.latest_dataset_list()
        latest_source_list = latest_sources.get("sources", [])
        if len(sources) == 0 and len(latest_source_list) == 0:
            sources_dict = {"sources": []}
            self.save(sources_dict)
        else:
            validated_sources = self._remove_invalid_dataset_formats(sources)
            if self.is_real_filter == True:
                validated_sources = self._filter_non_existing_datasets(
                    validated_sources
                )
            _sources = self._add_wo_duplicates(latest_source_list, validated_sources)
            sources_dict = {"sources": _sources}
            self.save(sources_dict)

    def count_dataset_list(self) -> int:
        """ Get the number of data source records we've gathered so far. """
        self.check()
        count = self.count()
        return count

    def latest_dataset_list(self) -> dict:
        self.check()
        latest_list = self.last()
        return latest_list


    

    def _load_dataset_list(self):
        self.check()
        if self.count_dataset_list() > 0:
            latest = self.latest_dataset_list()
            sources = latest.get("sources", [])
            if len(sources) == 0:
                return
            # print(sources)
            for source in sources:
                self._is_data_exist(source)

    def _reset_dataset_list(self) -> None:
        """ Initialize with a nil set of data if nothing exist yet"""
        if self.count_dataset_list() == 0:
            self.add_multiple_data_sources([], allow_bypass=True)

    # def dump(self)

    def reset(self):
        """ Reset the data we're querying for. """
        self._load_dataset_list()
        self._reset_dataset_list()
        self.time.reset()
        self.sync()
        self.metaid = self.metadata.reset()
        return self.metaid

    def step(self, call_type: str = "dataframe"):
        avail_types = ["dataframe", "current"]
        if call_type not in avail_types:
            call_type = "dataframe"

        data_set = {}
        # We can multi-thread this
        for dataset in self.datasets:
            dataset_name = str(dataset)
            dataset.event = self.event
            dataset.processor = self.processor
            dataset.preprocessor = self.preprocessor
            if call_type == "dataframe":
                data_set[dataset_name] = dataset.dataframe_from_head()
            else:
                data_set[dataset_name] = dataset.closest_head()
        """ Remove this time step """
        self.sync()
        return data_set
    

    def sync(self):
        """ Gets all of the datahandlers and synchronize their time object """
        if len(self.datasets) > 0:
            self.time.processor = self.processor
            for data in self.datasets:
                data.live = self.live
                data.episode = self.episode
                data.time = self.time
    

    def pick(self, _id:str):
        # Will probably make this kind of interface common. Get all information by meta_id
        """Pick

        Get multi-dataset by id

        Parameters
        ----------
        _id : str
            MultiData identifier
        """
        item = self.search.FindById(_id)
        if item is None:
            return None
        _multi                  = MultiDataManagement()
        _multi.processor        = self.processor
        _multi['name']          = item['name']
        _multi['abbreviation']  = item['abbreviation']
        _multi['submetatype']   = item['submetatype']
        _multi['category']      = item['category']
        _multi['metatype']      = item['metatype']
        _multi['subcategories'] = (item['subcategories']).to_dict()
        _multi.reset()
        return _multi

    

if __name__ == "__main__":
    with example_space("Multi-Data-Management") as example:
        # set_name = uuid.uuid4().hex
        set_name = "ac688d95336e41bdbe61c5c804d07f1a"
        # jam = Jamboree()
        jam_proc = JamboreeNew()
        multi_data = MultiDataManagement()
        multi_data["set_name"] = set_name
        # multi_data.event = jam
        multi_data.processor = jam_proc
        multi_data.episode = uuid.uuid4().hex
        multi_data.reset()
        dset1 = {
            "name": "shaw",
            "subcategories": {"beautiful": "mind", "it": "is"},
            "category": "pricing",
        }

        dset2 = {
            "name": "shank",
            "subcategories": {"wonderful": "hello", "king": "world"},
            "category": "pricing",
        }

        dset3 = {
            "name": "MSFT",
            "subcategories": {
                "market": "stock",
                "country": "US",
                "sector": "techologyyyyyyyy",
            },
            "category": "markets",
        }

        dset4 = {
            "name": "AAPL",
            "subcategories": {
                "market": "stock",
                "country": "US",
                "sector": "techologyyyyyyyy",
            },
            "category": "markets",
        }

        full_set = [dset1, dset2, dset3, dset4]
        pprint.pprint(multi_data.sources)
        multi_data.add_multiple_data_sources(full_set)
        # Check to make sure we aren't adding any dummy sources
        multi_data.time.head = maya.now().subtract(weeks=200, hours=14)._epoch
        multi_data.time.change_stepsize(microseconds=0, days=1, hours=0)
        multi_data.time.change_lookback(microseconds=0, weeks=4, hours=0)
        multi_data.sync()
        for _ in range(1000):
            multi_data.step()
            multi_data.time.step()
            current_time = multi_data.time.head
            print((f"Step {current_time}"))



"""
    Temporarily stores all keys that we'll possibly be using later.

    Entirely used to get the subdocuments by id. 
    Store all subdocuments by super_id


"""

from addict import Dict

class Keystore(object):
    def __init__(self):
        self.store = Dict()
    

    def add_by_superid(self, superid:str, key:str, _dict:dict):
        # _dict.pop("super_id", None) 
        _dict.pop("id", None)
        _dict.pop("payload", None)

        item = {
            key: _dict
        }
        super_item = {
            str(superid): item
        }
        self.store.update(super_item)
    
    def get_by_superid(self, superid:str):
        if superid in self.store:
            return self.store[superid]
        return {}
    

    def add(self, superid:str, key:str, _dict:dict):
        _dict.pop("super_id", None) 
        _dict.pop("id", None)
        _dict.pop("payload", None)

        item = {
            key: _dict
        }
        super_item = {
            str(superid): item
        }
        self.store.update(super_item)
    
    def get(self, superid:str):
        if superid in self.store:
            return self.store[superid]
        return {}
    
    def reset(self):
        self.store = Dict()


import re
from typing import List
from jamboree.utils.support.search import ( is_gen_type, is_generic, is_geo, is_valid_geo,
                                            is_nested, name_match, to_field,
                                            to_str, is_valid_text, is_valid_numeric, is_queryable_dict)



def boolean_process(field, is_true=False):
    """ Return a dicionary that has a TEXT value to represent a boolean """
    bstring = "FALSE"
    if is_true:
        bstring = "TRUE"
    return {
        field: bstring
    }

def list_process(field, item_list:List[str]):
    """ Return a dictionary representing a list of tags"""
    # if isinstance(item_list, list):
    if len(item_list) == 0:
        return {

        }
    saved_list = []

    for i in item_list:
        saved_list.append(f"{i}")
    return {
        field: ",".join(saved_list)
    }

def text_process(field:str, term:str, is_exact=False):
    if is_exact:
        return {
            field: term
        }
    return {
        field:term
    }

def number_process(field, number):
    return {
        field: number
    }



def geo_process_dict(field:str, dictionary:dict):
    """ Converts a dictionary into a dictionary string"""
    vals = dictionary['values']
    if not is_valid_geo(vals):
        return {}
    lon = vals.get("long")
    lat = vals.get("lat")
    return {
        field: f"{lon},{lat}"
    }

def num_process_dict(field:str, dictionary:dict):
    d_vals = dictionary['values']
    if is_valid_numeric(d_vals):
        _operation = d_vals.get("operation")
        _upper = d_vals.get("upper")
        _lower = d_vals.get("lower")
        
        if _operation == "greater":
            return number_process(field, _upper)
        elif _operation == "lesser":
            return number_process(field, _lower)
        elif _operation == "between":
            return number_process(field, _upper)
        elif _operation == "exact":
            _is_exact = (_upper == _lower)
            if _is_exact:
                return number_process(field, _upper)
        return {}

def text_process_dict(field, dictionary:dict):
    """ Create a simple text field from the dictionary"""
    values = dictionary.get("values")
    if is_valid_text(values):
        is_exact = values.get("is_exact", False)
        _term = values.get("term", False)
        filtered_term = re.sub('[^a-zA-Z0-9\n\.|\*|\@|\|\_]', ' ', _term)
        return text_process(field, filtered_term, is_exact=is_exact)
    return {

    }


def create_insertable(example:dict):
    insertable = {}
    for k, v in example.items():
        if isinstance(v, list):
            insertable.update(list_process(k, v))
        elif isinstance(v, str):
            insertable.update(text_process(k, v))
        elif isinstance(v, bool):
            insertable.update(boolean_process(k, v))
        elif isinstance(v, float) or isinstance(v, int):
            insertable.update(number_process(k, v))
        elif isinstance(v, dict):
            if not is_queryable_dict(v):
                continue
            if v['type'] == "NUMERIC":
                insertable.update(num_process_dict(k, v))
            
            if v['type'] == "GEO":
                insertable.update(geo_process_dict(k, v))
            
            if v['type'] == "TEXT":
                insertable.update(text_process_dict(k, v))
    return insertable


def main():
    """ Convert a dictionary into an insertable dictionary"""
    example = {
        "maybe": True,
        "gtags": ["one", "two", "three"],
        "current": {
            "type": "NUMERIC",
            "is_filter": True,
            "values": {
                "lower": 33,
                "upper": 0,
                "operation": "between"
            }
        },
        "loc": {
            "type": "GEO",
            "is_filter": True,
            "values": {
                "long": 33,
                "lat": -10,
                "distance": 1.2,
                "metric": "km"
            }
        },
        "exact_text": {
            "type": "TEXT",
            "is_filter": True,
            "values": {
                "term": "hello world"
            }
        },
    }
    

    for _ in range(100):
        insertable = create_insertable(example)
        
        print(insertable)
    pass


if __name__ == "__main__":
    main()


import maya
from threading import local
from jamboree.storage.files import FileStorageConnection
from jamboree.utils.core import consistent_hash
from jamboree.utils.support.storage import serialize, deserialize
from jamboree.utils.context import watch_loop
from addict import Dict
import redis
from redis import Redis
from redis.client import Pipeline
import version_query
from loguru import logger
logger.disable(__name__)

class RedisFileProcessor(object):
    def __init__(self, *args, **kwargs):
        self._pipe = None
        self._conn = None
    
    @property
    def conn(self) -> Redis:
        if self._conn is None:
            raise AttributeError("Pipe hasn't been set")
        return self._conn
    @conn.setter
    def conn(self, _pipe:Redis):
        self._conn = _pipe
    
    @property
    def pipe(self) -> Pipeline:
        if self._pipe is None:
            raise AttributeError("Pipe hasn't been set")
        return self._pipe
    @pipe.setter
    def pipe(self, _pipe:Pipeline):
        self._pipe = _pipe
    
    
    def reset(self):
        self.pipe = None
        self.conn = None


class RedisFileConnection(FileStorageConnection):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        """ NOTE: Expiriment with sorted sets """
        self.current_query = {}
        self.current_hash = None
        self.current_query_exist = None
        self.current_pipe = None
        self.current_hash_keys = None
        self.current_version = None
        self.current_version_exist = None
        self.setup_run = None

    def gwatch(self):
        sorted_version = self.keys.version.sorted
        set_version = self.keys.version.set
        self.pipe.watch(sorted_version)
        self.pipe.watch(set_version)

    @property
    def version(self):
        """ Get the latest version or the default"""
        sorted_version = self.keys.version.sorted
        set_version = self.keys.version.set
        # self.pipe.watch(sorted_version)
        # self.pipe.watch(set_version)
        if self.query_exists and self.current_version is None:
            # latest_version = self.connection.zrange(sorted_version, -1, -1)
            # _all_versions = self.connection.zrange(sorted_version, 0, -1)
            latest_version = self.pipe.zrange(sorted_version, -1, -1)
            
            if latest_version is not None and len(latest_version) > 0:
                latest_version = latest_version[0]
                self.current_version = latest_version.decode()
        elif self.current_version is not None:
            return self.current_version
        else:
            latest_version = self.settings.default.version
            self.pipe.zadd(sorted_version, {latest_version: maya.now()._epoch})
            self.pipe.sadd(set_version, latest_version)
            self.current_version = latest_version
        return self.current_version
    
    @version.setter
    def version(self, _version:str):
        sorted_version = self.keys.version.sorted
        set_version = self.keys.version.set
        self.pipe.zadd(sorted_version, {_version: maya.now()._epoch})
        self.pipe.sadd(set_version, *_version)
        self.current_version = _version
    


    @property
    def hash_query(self):
        if self.current_hash is None:
            self.current_hash = consistent_hash(self.current_query)
        return self.current_hash
    
    @property
    def keys(self):
        """ Set all of the required keys to something that fits in memory. """
        if self.current_hash_keys is None:
            self.current_hash_keys = Dict()
            self.current_hash_keys.version.set = f"{self.hash_query}:versions"
            self.current_hash_keys.version.sorted = f"{self.hash_query}:zversions"
            self.current_hash_keys.file.sum = f"{self.hash_query}:sums"
            self.current_hash_keys.sum = f"{self.hash_query}:sum"
            self.current_hash_keys.version.index = f"{self.hash_query}:incr"
        return self.current_hash_keys
    
    @property
    def query_exists(self) -> bool:
        if self.current_query_exist is None:
            version_set_exist = self.pipe.exists(self.keys.version.set)
            sorted_version_exist = self.pipe.exists(self.keys.version.sorted)

            self.current_query_exist = (version_set_exist == 1 and sorted_version_exist == 1)
        return self.current_query_exist
    
    @property
    def file_exist(self) -> bool:
        """ Does the current file version exist"""
        
        if self.current_version_exist is None:
            vk = self.version_key
            self.pipe.watch(vk)
            version_set_exist = self.pipe.exists(vk)
            self.current_version_exist = (version_set_exist == 1)
        return self.current_version_exist
    
    @property
    def pipe(self):
        if self.current_pipe is None:
            raise AttributeError("Pipe cannot be non-existent")
        return self.current_pipe

    @property
    def version_key(self) -> str:
        return f"{self.hash_query}:{self.version}"
    

    def update_version(self):
        """ Save version in multiple places to be found later"""
        version =  self.version
        if self.query_exists and not self.is_overwrite:
            vs = version_query.Version.from_str(version)
            new_vs = vs.increment(self.settings.default.increment)
            new_vs_str = new_vs.to_str()
            # print(new_vs_str)
            self.version = new_vs_str
        
    
    def update_file(self, _file):
        """Update the file"""
        vkey = self.version_key
        self.pipe.set(vkey, _file)

    def update(self, file):
        """ Update the file version and update the file. """
        self.update_version()
        self.update_file(file)
        logger.error(file)

    @property
    def garbage_patch(self):
        """ 
            Basically a way to see if both the file and query key exist. 
            It's pretty janky. It'll be fine though. 
        """
        return (self.query_exists and self.file_exist)

    def absolute_exists(self, query, **kwargs):
        # self.setup_run = None
        self.setup(query, **kwargs)
        return (self.query_exists and self.file_exist)


    @logger.catch
    def save(self, query:dict, obj, **kwargs):
        self.setup(query, **kwargs)
        serial_item = serialize(obj)
        self.update(serial_item)
        
        
    

    @logger.catch
    def query(self, query:dict, **kwargs):
        self.setup(query, **kwargs)
        
        if self.garbage_patch:
            # If the query and file exist
            # logger.debug("File exist, we're gonna try pulling it")
            # logger.debug(self.version)
            item = self.pipe.get(self.version_key)
            unpacked = deserialize(item)
            if unpacked is None:
                raise AttributeError("Pickled Item Not Found")
            return unpacked


    def delete(self, query:dict, **kwargs):
        self.setup(query, **kwargs)
        if self.query_exists and self.file_exist:
            # If the query and file exist
            sorted_version = self.keys.version.sorted
            set_version = self.keys.version.set
            self.pipe.delete(self.version_key)
            self.pipe.zrem(sorted_version, self.version)
            self.pipe.srem(set_version, self.version)
    
    def setup(self, query:dict, **kwargs):
        is_force = kwargs.pop("is_force", False)
        if self.setup_run is None or is_force:
            self.reset()
            self.settings = Dict(**kwargs)
            self.current_query = query
            self.gwatch()
            self.version
            self.setup_run = True
            self.file_exist
    
    def reset(self):
        """ Reset all placeholder variables"""
        self.current_query = {}
        self.current_hash = None
        self.current_query_exist = None
        self.current_pipe = self.conn.pipeline()
        self.current_version = None
        self.current_version_exist = None
        self.setup_run = None



class SampleObj(object):
    def __init__(self) -> None:
        self.one = "one"
        self.two = "two"

def main():
    current_settings = Dict()
    query_dict = Dict()
    current_settings.overwrite = True
    current_settings.preferences = query_dict
    samp_opt = SampleObj()
    redpill = redis.Redis()
    redconn = RedisFileConnection()
    redconn.conn = redpill
    redconn.save({"one": "twoss"}, samp_opt, **current_settings)
    item = redconn.query({"one": "twoss"})
    logger.info(item)

    # redconn.pipe.execute()

if __name__ == "__main__":
    main()


import uuid
import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)
import maya
import pprint
from loguru import logger
from typing import List
from jamboree.handlers.default import DataHandler
from jamboree import Jamboree
from jamboree.utils.support.search import querying
# from jamboree.handlers.abstracted.search.meta import MetaDataSearchHandler
class OrderbookData(DataHandler):
    """
        # Orderbook Data
        
        A way to browse and interact with price data. Is an extension of DataHandler and includes basic searches.

    """
    def __init__(self):
        super().__init__()
        self['subtype'] = "orderbook" 
        self['category'] = "markets"

        self.sc = "subcategories" # storing the placeholder key to prevent misspelling
        self.cat = "category" # storing variable placeholder key to prevent misspelling

    @property
    def markets(self) -> List[str]:
        return [
            'crypto', 'stock', 'commodities', 'forex', 'simulation'
        ]


    def by_market(self, market_type:str):
        """
            # Find All Datasets By Market

            market_type: ['crypto', 'stock', 'commodities', 'forex', 'simulation']
        """
        if market_type not in self.markets:
            logger.error(f"Not the correct type: {market_type} must be {self.markets}")
            return []

        _search = self.search
        _search[self.cat] = "markets"
        _search[self.sc] = {
            "market": market_type
        }
        # print(_search.query_builder.build())
        return _search.find()
    
    def by_country(self, country:str):

        """
            # Find All Datasets By Country

        """

        if not isinstance(country, str):
            logger.error("The country is not the string.")
            return []
        _search = self.search
        _search[self.cat] = "markets"
        _search[self.sc] = {
            "country": country
        }
        return _search.find()
    
    def by_sector(self, sector:str):
        """
            # Find All Datasets By Sector

        """


        if not isinstance(sector, str):
            logger.error("The sector should be a string.")
            return []
        _search = self.search
        _search[self.cat] = "markets"
        _search[self.sc] = {
            "sector": sector
        }
        return _search.find()
    
    def by_name(self, name:str):
        """
            # Find All Datasets By Sector

        """


        if not isinstance(name, str):
            logger.error("The sector should be a string.")
            return []
        _search = self.search
        _search[self.cat] = "markets"
        _search["name"] = name
        return _search.find()
    
    
    
    def multi_part_search(self, name=None, country=None, sector=None, market=None, exchange=None):
        """  """
        all_variables = {"name": name, "country": country, "sector": sector, "market": market, "exchange": exchange}
        _name = None
        _subcat_dict = {}
        for k, v in all_variables.items():
            if v is None:
                continue
            if k == "name":
                _name = v
                continue

            _subcat_dict[k] = querying.text.exact(v)
        
        is_size = (len(_subcat_dict) == 0)
        is_name = (_name is None)
        if is_size and is_name:
            return []

        _search = self.search
        # _search.reset()

        if not is_name:
            logger.warning(name)
            _search["name"] = name
        _search[self.cat] = "markets"
        if not is_size:
            _search[self.sc] = _subcat_dict
        _search.processor = self.processor
        return _search.find()

def main():
    import pandas_datareader.data as web
    # data_msft = web.DataReader('MSFT','yahoo',start='2019/9/1',end='2020/1/30').round(2)
    # data_apple = web.DataReader('AAPL','yahoo',start='2019/9/1',end='2020/1/30').round(2)
    episode_id = uuid.uuid4().hex
    jambo = Jamboree()
    jam_processor = Jamboree()
    data_hander = OrderbookData()
    data_hander.processor = jam_processor
    data_hander.event = jambo
    # The episode and live parameters are probably not good for the scenario. Will probably need to switch to something else to identify data
    data_hander.episode = episode_id
    data_hander.live = False
    data_hander['subcategories'] = {
        "market": "stock",
        "country": "Mexico",
        "sector": "tech",
        "exchange": "binance",
    }
    data_hander['name'] = "ETH Ethereum"
    data_hander.reset()
    # data_hander.store_time_df(data_msft, is_bar=True)


    data_hander['name'] = "BTC Bitcoin"
    # data_hander.reset()
    # data_hander.store_time_df(data_apple, is_bar=True)

    start = maya.now()._epoch
    # res1 = data_hander.by_name("Bitcoin")

    # logger.debug(res1)
    end = maya.now()._epoch
    logger.info(end-start)

    res = data_hander.multi_part_search(market="stock", exchange="binance", country="Mexico")
    logger.warning(res)
    
    # _search = data_hander.search
    # _search['subcategories'] = {
    #     "market": "stock"
    # }
    # _search.remove()
    # data_hander.time.head = maya.now().subtract(weeks=200, hours=14)._epoch
    # data_hander.time.change_stepsize(microseconds=0, days=1, hours=0)
    # data_hander.time.change_lookback(microseconds=0, weeks=4, hours=0)

    
    # while data_hander.is_next:
    #     logger.debug(data_hander.time.head)
    #     print(data_hander.closest_head())
    #     data_hander.time.step()


if __name__ == "__main__":
    main()


import uuid
import time
import maya
from jamboree.handlers.default import DBHandler
from jamboree import Jamboree
from jamboree import JamboreeNew
from loguru import logger


class TimeHandler(DBHandler):
    """ 
        # TimeHandler
        --- 
        The time handler is used to both do simple calculations of time and to maintain the head.

        The `head` is the current time we're referring to. 

        ```
            time_handler = TimeHandler()
            time_handler['episode'] = "episodeid"
            time_handler['live'] = False
            time_handler.reset()
            current_time:float = time_handler.head
        ```
    """

    def __init__(self):
        super().__init__()
        self.entity = "timeindexing"
        self.required = {
            "episode": str,
            "live": bool,
        }
        self.micro = 1000
        self.seconds = 0
        self.minutes = 0
        self.hours = 10
        self.days = 0
        self.weeks = 0

        self.step_micro = 1000
        self.step_seconds = 0
        self.step_minutes = 0
        self.step_hours = 10
        self.step_days = 0
        self.step_weeks = 0

        self._head = time.time()

    @property
    def looks(self):
        lookback_dict = {
            "microseconds": self.micro,
            "seconds": self.seconds,
            "minutes": self.minutes,
            "hours": self.hours,
            "days": self.days,
            "weeks": self.weeks,
        }
        return lookback_dict

    @property
    def steps(self):
        stepsize_dict = {
            "microseconds": self.step_micro,
            "seconds": self.step_seconds,
            "minutes": self.step_minutes,
            "hours": self.step_hours,
            "days": self.step_days,
            "weeks": self.step_weeks,
        }
        return stepsize_dict

    @property
    def lookback_params(self):
        self.load_lookback()
        lookback_dict = self.looks
        return lookback_dict

    @property
    def stepsize_params(self):
        self.load_stepsize()
        stepsize_dict = self.steps
        return stepsize_dict

    @property
    def head(self):
        if self["episode"] == "live" and self["live"] == True:
            return maya.now()._epoch
        self.load_head()
        return self._head

    @head.setter
    def head(self, _head):

        self._head = _head
        self.save_head()

    @property
    def tail(self):
        head = self.head
        look_params = self.lookback_params
        tail = maya.MayaDT(head).subtract(**look_params)
        return tail._epoch

    def change_lookback(
        self, microseconds=1000, seconds=0, minutes=0, hours=10, days=0, weeks=0
    ):
        self.load_lookback()
        if microseconds != 1000:
            self.micro = microseconds
        if seconds != 0:
            self.seconds = seconds
        if minutes != 0:
            self.minutes = minutes
        if hours != 10:
            self.hours = hours
        if days != 0:
            self.days = days
        if weeks != 0:
            self.weeks = weeks

        self.save_lookback()

    def change_stepsize(
        self, microseconds=1000, seconds=0, minutes=0, hours=10, days=0, weeks=0
    ):
        self.load_stepsize()
        if microseconds != 1000:
            self.step_micro = microseconds
        if seconds != 0:
            self.step_seconds = seconds
        if minutes != 0:
            self.step_minutes = minutes
        if hours != 10:
            self.step_hours = hours
        if days != 0:
            self.step_days = days
        if weeks != 0:
            self.step_weeks = weeks

        self.save_stepsize()

    """ 
        # Index Commands
    """

    def save_head(self):
        alt = {"detail": "head"}
        # head_data = {"time": self._head}
        self.set_single(self._head, alt=alt, is_serialized=False)

    def latest_head(self) -> dict:
        alt = {"detail": "head"}
        item = self.get_single(alt=alt, is_serialized=False)
        return item

    # def

    def count_headindex(self) -> int:
        alt = {"detail": "headindex"}
        lookback_count = self.count(alt=alt)
        return lookback_count

    def save_headindex(self):
        """ Save monitored assets """
        alt = {"detail": "headindex"}
        self.save({"message": "Head Index"}, alt=alt)

    """ 
        # Lookback Commands
    """

    def count_lookback(self) -> int:
        alt = {"detail": "lookback"}
        lookback_count = self.count(alt=alt)
        return lookback_count

    def latest_lookback(self) -> dict:
        alt = {"detail": "lookback"}
        lookback = self.last(alt=alt)
        return lookback

    def save_lookback(self):
        """ Save monitored assets """
        alt = {"detail": "lookback"}
        params = self.looks
        self.save(params, alt=alt)


    def count_stepsize(self) -> int:
        alt = {"detail": "stepsize"}
        lookback_count = self.count(alt=alt)
        return lookback_count

    def latest_stepsize(self) -> dict:
        alt = {"detail": "stepsize"}
        lookback = self.last(alt=alt)
        return lookback

    def save_stepsize(self):
        """ Save monitored assets """
        alt = {"detail": "stepsize"}
        self.save(self.steps, alt=alt)
    def load_lookback(self):
        count = self.count_lookback()
        if count > 0:
            look = self.latest_lookback()
            self.micro = look.get("microseconds", 1000)
            self.seconds = look.get("seconds", 0)
            self.minutes = look.get("minutes", 0)
            self.hours = look.get("hours", 10)
            self.days = look.get("days", 0)
            self.weeks = look.get("weeks", 0)

    def load_stepsize(self):
        count = self.count_stepsize()
        if count > 0:
            look = self.latest_stepsize()
            self.step_micro = look.get("microseconds", 1000)
            self.step_seconds = look.get("seconds", 0)
            self.step_minutes = look.get("minutes", 0)
            self.step_hours = look.get("hours", 10)
            self.step_days = look.get("days", 0)
            self.step_weeks = look.get("weeks", 0)

    def load_head(self):
        if self.count_headindex() > 0:
            head = self.latest_head()
            if head is None:
                return time.time()
            self._head = float(head.decode("utf-8"))

    def _reset_lookback(self):
        if self.count_lookback() == 0:
            self.save_lookback()

    def _reset_headindex(self):
        if self.count_headindex() == 0:
            self.save_headindex()
            self.save_head()

    def reset(self):
        self._reset_lookback()
        self._reset_headindex()

    def step(self) -> None:
        head = self.head
        step_params = self.stepsize_params
        self.head = maya.MayaDT(head).add(**step_params)._epoch
        self.save_head()

    def peak(self) -> float:
        """ Get the time exactly one step ahead"""
        head = self.head
        step_params = self.stepsize_params
        new_head = maya.MayaDT(head).add(**step_params)._epoch
        return new_head

    def peak_far(self) -> float:
        """ 
            # PEAK FAR
            ---
            Peak far into the future. One step head + one lookahead_params 

            Doesn't modify the head
        """
        head = self.head
        step_params = self.stepsize_params
        lookahead_params = self.lookback_params
        new_head = maya.MayaDT(head).add(**step_params).add(**lookahead_params)._epoch
        return new_head

    def peak_back(self) -> float:
        head = self.head
        step_params = self.stepsize_params
        new_head = maya.MayaDT(head).subtract(**step_params)._epoch
        return new_head

    def peak_back_far(self) -> float:
        """ 
            # PEAK FAR
            ---
            Peak far into the future. One step head + one lookahead_params 
        """
        head = self.head
        step_params = self.stepsize_params
        lookahead_params = self.lookback_params
        new_head = (
            maya.MayaDT(head)
            .subtract(**step_params)
            .subtract(**lookahead_params)
            ._epoch
        )
        return new_head

    def peak_back_num(self, n: int = 1) -> float:
        """ 
            # Peak Back Num
            ---
            Peak far into the future. One step head + one lookahead_params

            Parameters
            ----------
                n: {int} - The number of steps backwards. Used to get the information.
        """
        if n < 1:
            raise ValueError("Number n needs to be greater than 1")
        head = self.head
        step_params = self.stepsize_params
        current_position = maya.MayaDT(head)
        for _ in range(n):
            current_position = current_position.subtract(**step_params)
        peak_position = current_position._epoch
        return peak_position

    def peak_back_num_tail(self, n: int = 1) -> float:
        """ 
            # Peak Back Num
            ---
            Peak far into the future. One step head + one lookahead_params

            Parameters
            ----------
                n: {int} - The number of steps backwards. Used to get the information.
        """
        if n < 1:
            raise ValueError("Number n needs to be greater than 1")
        head = self.head
        step_params = self.stepsize_params
        look_params = self.lookback_params
        current_position = maya.MayaDT(head)
        for _ in range(n):
            current_position = current_position.subtract(**step_params)
        peak_position = current_position.subtract(**look_params)._epoch
        return peak_position

    def step_back(self) -> None:
        head = self.head
        step_params = self.stepsize_params
        self.head = maya.MayaDT(head).subtract(**step_params)._epoch
        self.save_head()


if __name__ == "__main__":
    jambo = Jamboree()
    timehandler = TimeHandler()
    timehandler.processor = jambo
    eid = uuid.uuid4().hex
    timehandler["episode"] = eid
    timehandler["live"] = False
    timehandler.reset()
    
    timehandler.head = maya.MayaDT(timehandler.head).subtract(weeks=20, days=9)._epoch
    timehandler.change_stepsize(microseconds=0, days=1, hours=0)
    for _ in range(100):
        logger.warning(maya.MayaDT(timehandler.head))
        timehandler.step()


from copy import copy
from abc import ABC
from typing import Any, List
from addict import Dict
from version_query import VersionComponent
import hashlib 

class FileStorageConnection(ABC):
    def __init__(self, **kwargs) -> None:
        self._connection = None
        self._settings = Dict()
        self._settings.overwrite = False
        self._settings.sig_key = kwargs.get("signature", "basic_key")
        
        self._settings.preferences.by = "latest"
        self._settings.preferences.limit = 500
        self._settings.preferences.version = None
        self._settings.default.version = "0.0.1"
        self._settings.default.increment = VersionComponent.Patch
        



    @property
    def conn(self):
        if self._connection is None:
            raise AttributeError("You haven't added a main database connection as of yet.")
        return self._connection

    @conn.setter
    def conn(self, _conn):
        self._connection = _conn

    @property
    def settings(self):
        return self._settings

    @settings.setter
    def settings(self, _settings:Dict):
        copied = self._settings
        copied.update(_settings)
        self.valid_settings(copied)
        self._settings = copied

    def valid_settings(self, _settings):
        if self._settings.preferences.by not in ["latest", "many", "all", "version"]:
            raise ValueError("We must query within a given range of types: 'latest', 'all', 'version'")
        
        if self._settings.preferences.by == "version" and self._settings.preferences.version is None:
            raise AttributeError("If you're querying by version, you have to include a version number (string_format)")

    @property
    def is_overwrite(self) -> bool:
        return self._settings.overwrite


    """ Main Commands """

    def save(self, query:dict, obj:Any, **kwargs):
        raise NotImplementedError("save not implemented")
    
    def query(self, query, **kwargs):
        raise NotImplementedError("query not implemented")
    
    def delete(self, query, **kwargs):
        raise NotImplementedError("delete_latest not implemented")

    def absolute_exists(self, query, **kwargs):
        raise NotImplementedError


import uuid
from typing import Any, AnyStr, Dict

import maya
import pandas as pd
import ujson
from jamboree import Jamboree, JamboreeNew
from jamboree.handlers.abstracted.search import MetadataSearchHandler
from jamboree.handlers.complex.meta import MetaHandler
from jamboree.handlers.default.db import DBHandler
from jamboree.handlers.default.time import TimeHandler
from jamboree.handlers.processors import (DataProcessorsAbstract,
                                          DynamicResample)
from jamboree.utils import omit
from jamboree.utils.support.search import querying
from loguru import logger


class Access(DBHandler):
    # ---------------------------------------------------------------------------------
    #                          Simple Accessor Properties
    # ---------------------------------------------------------------------------------
    
    @property
    def name(self):
        return self['name']

    @name.setter
    def name(self, __name: str):
        self['name'] = __name

    
    @property
    def category(self) -> str:
        return self['category']

    @category.setter
    def category(self, _category: str):
        self['category'] = _category

    @property
    def subcategories(self) -> str:
        return self['subcategories']

    @subcategories.setter
    def subcategories(self, __subcatgeories: Dict[AnyStr, Any]):
        self['subcategories'] = __subcatgeories

    @property
    def metatype(self) -> str:
        return self['metatype']

    @metatype.setter
    def metatype(self, __metatype: str):
        self['metatype'] = __metatype

    @property
    def submetatype(self) -> str:
        return self['submetatype']

    @submetatype.setter
    def submetatype(self, __submetatype: str):
        self['submetatype'] = __submetatype

    @property
    def abbreviation(self) -> str:
        return self['abbreviation']

    @abbreviation.setter
    def abbreviation(self, __abb: str):
        self['abbreviation'] = __abb



import copy
import inspect
import pprint
from typing import Any, Dict, Optional, AnyStr

import ujson
from loguru import logger

from jamboree import Jamboree, JamboreeNew
from jamboree.base.processors.abstracts import EventProcessor, Processor
from jamboree.handlers.base import BaseHandler
from jamboree.handlers.default.search import BaseSearchHandler
from jamboree.utils import memoized_method
from jamboree.utils.helper import Helpers


class DBHandler(BaseHandler):
    """ 
        A simple event store using a variation of databases.
        ---
        
        Currently uses zadd to work
    """

    def __init__(self):
        # print("DBHandler")
        self._metatype = "event"
        self._entity = ""
        self._required = {}
        self._query = {}
        self.data = {}
        self._is_event = True
        self._processor: Optional[Processor] = None
        self.event_proc: Optional[EventProcessor] = None
        self.main_helper = Helpers()

    def __setitem__(self, key, value):
        if bool(self.required):
            if key in self.required:
                self._query[key] = value
                return self._query
        self.data[key] = value
        return self.data

    def __getitem__(self, key):
        if key in self._query.keys():
            return self._query.get(key, None)
        else:
            if key in self.data.keys():
                return self.data.get(key, None)
        return None
    
    def setup_query(self, alt={}):
        return self.setup_key_with_lru(self._query, self.entity, self._metatype, self.data, alt)

    def setup_key_with_lru(self, query:dict, entity:str, metatype:str, data:dict, alt:dict):
        query = copy.copy(query)
        query['type'] = entity
        query['mtype'] = metatype
        query.update(alt)
        query.update(data)
        return query


    @property
    def is_event(self) -> bool:
        """ Determines if we're going to add event ids to what we're doing. We can essentially set certain conditions"""
        return self._is_event
    
    @is_event.setter
    def is_event(self, is_true:bool=False):
        self._is_event = is_true

    @property
    def processor(self) -> Processor:
        if self._processor is None:
            raise AttributeError("The Processor is missing")
        return self._processor
    
    @processor.setter
    def processor(self, _processor: 'Processor'):
        self._processor = _processor


    @property
    def event(self) -> Optional['EventProcessor'] :
        return self.event_proc

    @event.setter
    def event(self, _event: EventProcessor):
        # Use to process event
        self.event_proc = _event
    
    def clear_event(self) -> None:
        self.event_proc = None
        self._processor = None

    @property
    def entity(self):
        return self._entity

    @entity.setter
    def entity(self, _entity: str):
        self._entity = str(_entity)

    @property
    def required(self):
        return self._required

    @required.setter
    def required(self, _required: Dict[str, Any]):
        # check to make sure it's not empty
        self._required = _required

    @property
    def query(self):
        return self._query

    @query.setter
    def query(self, _query: Dict[str, Any]):
        if len(_query.keys()) > 0:
            self._query = _query

    def check(self):
        
        # self.processor
        # if self.event_proc is None:
        #     raise AttributeError("Event processor isn't available.")
        if (not bool(self._entity)):
            raise AttributeError("Entity hasn't been set")
        
        if (not bool(self._required)):
            raise AttributeError("None of the required information has been set")
        
        if (not bool(self._query)):
            raise AttributeError("None of the queryable information has been set")
        

        for req in self._required.keys():
            _type = self._required[req]
            if req not in self._query:
                raise AttributeError(f"{req} is not in the requirements")
            if not isinstance(self._query[req], _type):
                raise AttributeError(f"{req} is not a {_type}")
        return True

    

    def _get_many(self, limit: int, ar:str, alt={}):
        """ Aims to get many variables """
        query = self.setup_query(alt)
        latest_many = self.processor.event.get_latest_many(query, abs_rel=ar, limit=limit)
        return latest_many

    def _get_latest(self, ar, alt={}):
        query = self.setup_query(alt)
        
        latest = self.processor.event.get_latest(query, abs_rel=ar)
        return latest


    def _last_by(self, time_index:float, ar="absolute", alt={}) -> dict:
        query = self.setup_query(alt)
        return self.processor.event.get_latest_by(query, time_index, abs_rel=ar)


    def _in_between(self, min_epoch:float, max_epoch:float, ar:str="absolute", alt={}):
        query = self.setup_query(alt)
        return self.processor.event.get_between(query, min_epoch, max_epoch, abs_rel=ar)


    def save(self, data: dict, alt={}):
        self.check()
        query = self.setup_query(alt)
        if self.is_event:
            data = self.main_helper.add_event_id(data)
        self.processor.event.save(query, data)

    def save_many(self, data: list, ar="absolute", alt={}):
        self.check()
        if not self.main_helper.is_abs_rel(ar): return
        query = self.setup_query(alt)
        if self.is_event:
            data = self.main_helper.add_event_ids(data)
        self.processor.event.save_many(query, data)


    def last(self, ar="absolute", alt={}):
        self.check()
        if not self.main_helper.is_abs_rel(ar): return {}
        alt.update(self.data)
        return self._get_latest(ar=ar, alt=alt)

    
    def many(self, limit=1000, ar="absolute", alt={}):
        self.check()
        if not self.main_helper.is_abs_rel(ar): return []
        alt.update(self.data)
        return self._get_many(limit, ar, alt=alt)


    def pop(self, alt={}):
        self.check()
        query = self.setup_query(alt)
        self.processor.event.remove_first(query)


    def pop_many(self, _limit, alt={}):
        self.check()
        query = self.setup_query(alt)
        return self.processor.event.pop_multiple(query, _limit)


    
    def last_by(self, time_index:float, ar="absolute", alt={}):
        self.check()
        if not self.main_helper.is_abs_rel(ar): return {}
        item = self._last_by(time_index, ar=ar, alt=alt)
        return item    


    def in_between(self, min_epoch:float, max_epoch:float, ar:str="absolute", alt={}):
        self.check()
        if not self.main_helper.is_abs_rel(ar): return []
        items = self._in_between(min_epoch, max_epoch, ar=ar, alt=alt)
        return items



    def count(self, alt={}) -> int:
        """ Aims to get many variables """
        self.check()
        query = self.setup_query(alt)
        return self.processor.event.count(query)
    

    def get_single(self, alt={}, is_serialized=True):
        self.check()
        query = self.setup_query(alt)
        item = self.processor.event.single_get(query, is_serialized=is_serialized)
        return item

    def set_single(self, data:dict, alt={}, is_serialized=True):
        self.check()
        query = self.setup_query(alt)
        self.processor.event.single_set(query, data, is_serialized=is_serialized)

    def delete_single(self, alt={}, is_dumps=False):
        self.check()
        query = self.setup_query(alt)
        self.processor.event.single_delete(query)
    
    def delete_all(self, alt={}):
        self.check()
        query = self.setup_query(alt)
        self.processor.event.delete_all(query)

    def query_all(self, alt:Dict[AnyStr, Any]={}):
        self.check()
        query = self.setup_query(alt)
        items = self.processor.event.get_all(query)
        return items

    def get_minimum_time(self, alt:Dict[AnyStr, Any]={}):
        self.check()
        _query = self.setup_query(alt)
        _time = self.processor.event.min_time(_query)
        return _time

    def get_maximum_time(self, alt:Dict[AnyStr, Any]={}):
        self.check()
        _query = self.setup_query(alt)
        _time = self.processor.event.max_time(_query)
        return _time

    def copy(self):
        """ Get everything about this DBHandler without the event inside """
        # _event = self.event
        # self.event = _event
        current_dict = copy.copy(self.__dict__)
        non_lock_types = {

        }
        _process = self.processor
        self._processor = None
        for key, value in current_dict.items():
            classified = type(value)
            if inspect.isclass(classified):
                if isinstance(value, BaseHandler) or issubclass(classified, BaseHandler) or issubclass(classified, BaseSearchHandler):
                    non_lock_types[key] = value
                    setattr(self, key, None)

        

        self.clear_event()
        copied:self = copy.deepcopy(self)
        copied.processor = _process
        for k, v in non_lock_types.items():
            setattr(copied, k, v)

        self.__dict__ = current_dict
        self.processor = _process
        return copied
    
    def lock(self, alt={}):
        self.check()
        query = self.setup_query(alt)
        return self.processor.event.lock(query)



import maya
import orjson
from copy import copy

from redis.client import Pipeline
from loguru import logger
from typing import Dict, List, Optional, Any, AnyStr
from pprint import pprint
from jamboree.storage.databases import DatabaseConnection

# NOTE: Can add pipelining to redis storage to make fewer calls.


class RedisDatabaseZSetsConnection(DatabaseConnection):
    """ 
        Redis Connection Notes
        ---
        We're going to end up using zadds and zscores. 
        The indexable score makes it a good target for time series. 
        The fact that it's a set makes it so there can be no duplicates added.

        In essense. Perfect for time series. 
    """

    def __init__(self) -> None:
        super().__init__()

    """
        # Individual Access Methods
        ---
        Use commands here to have single variable in place instead of creating an event log. 
        Very useful for stateful applications such as timeindexing logic.

        ## Core Functions:

        * kill - delete a given hash/key
        * add  - set a given hash/key to a variable. Usually that should be a dictionary representing a complex value.
        * get  - get a variable by hash/key
    """

    def _kill(self, _hash: str):
        """ Deletes a key within a lock"""
        rlock = f"{_hash}:lock"
        sub_key = f"{_hash}:single"
        with self.connection.pipeline() as pipe:
            with pipe.lock(rlock):
                pipe.delete(sub_key)
                pipe.execute()

    def _add(self, _hash: str, data: dict):
        rlock = f"{_hash}:lock"
        sub_key = f"{_hash}:single"
        serialized = orjson.dumps(data)
        with self.connection.pipeline() as pipe:
            with pipe.lock(rlock):
                pipe.set(sub_key, serialized)
                pipe.execute()

    def _get(self, _hash: str):
        rlock = f"{_hash}:lock"
        sub_key = f"{_hash}:single"
        value = None
        with self.connection.pipeline() as pipe:
            with pipe.lock(rlock):
                pipe.watch(sub_key)
                value = pipe.get(sub_key)
                pipe.execute()
        return value

    def get(self, query: dict):
        if not self.helpers.validate_query(query):
            return {}

        _hash = self.helpers.generate_hash(query)
        value = self._get(_hash)
        if value is not None:
            return orjson.loads(value)
        return {}

    def add(self, query: dict, data: dict):
        """ Sets a single value. It's a wrapper around"""
        if not self.helpers.validate_query(query) or len(data) == 0:
            return

        _hash = self.helpers.generate_hash(query)
        self._add(_hash, data)

    def kill(self, query: dict):
        """ Deletes a single variable. Bypasses the stack"""
        if not self.helpers.validate_query(query):
            return

        _hash = self.helpers.generate_hash(query)
        self._kill(_hash)

    """ 
        # Save Commands
        ---
        * `save` - ...
        * `save_many` - ...
    """

    def _save(self, _hash: str, data: dict, timing: dict):
        """ Appends an event to the stack. """
        serialized = orjson.dumps(data)
        rlock = f"{_hash}:lock"
        relative_time_key = f"{_hash}:rlist"
        absolute_time_key = f"{_hash}:alist"
        with self.connection.pipeline() as pipe:
            with pipe.lock(rlock):
                relative_data = {serialized: timing["time"]}
                absolute_data = {serialized: timing["timestamp"]}
                pipe.zadd(relative_time_key, relative_data)
                pipe.zadd(absolute_time_key, absolute_data)
                pipe.execute()

    def save(self, query: dict, data: dict, _time=None, _timestamp=None):
        """ Save a single record. """
        if not self.helpers.validate_query(query):
            return
        _hash = self.helpers.generate_hash(query)
        query.update(data)
        data, timing = self.helpers.separate_time_data(query, _time, _timestamp)
        self._save(_hash, data, timing)

    def _save_many(self, _hash: str, relative_data: Dict[str, float] = {}):
        # serialized_list = [orjson.dumps(x) for x in data]

        rlock = f"{_hash}:lock"
        relative_time_key = f"{_hash}:rlist"
        absolute_time_key = f"{_hash}:alist"
        absolute_data = self.helpers.get_current_abs_time(relative_data)
        with self.connection.pipeline() as pipe:
            with pipe.lock(rlock):
                pipe.zadd(relative_time_key, relative_data)
                pipe.zadd(absolute_time_key, absolute_data)
                pipe.execute()

    def save_many(self, query, data: Dict[str, float] = {}, abs_rel="absolute"):
        """ 
            # Save Many

            Save multiple records at once. We assume the data was already processed. 
        """
        if not self.helpers.validate_query(query) or len(data) == 0:
            return

        _hash = self.helpers.generate_hash(query)
        self._save_many(_hash, data)

    """ 
        # Delete Commands
        ---
        * `delete` - Get all of the records related to a given key.
        * `delete_latest` - Get the `n` latest records according to our query parameters.
        * `delete_between` - Query between two epoch times.
        * `delete_before` - Get everything before an epoch time.
        * `delete_after` - Get everything after epoch time.

    """

    def delete_first(self, query, details):
        pass

    def _delete(self, _hash: str, details: Dict[str, Any]):
        rlock = f"{_hash}:lock"
        relative_time_key = f"{_hash}:rlist"
        absolute_time_key = f"{_hash}:alist"
        deletion_key = orjson.dumps(details)
        with self.connection.lock(rlock):
            self.connection.zrem(relative_time_key, deletion_key)
            self.connection.zrem(absolute_time_key, deletion_key)

    def delete(self, query: Dict[str, Any], details: Dict[str, Any]):
        if not self.helpers.validate_query(query):
            return
        _hash = self.helpers.generate_hash(query)
        count = self.count(_hash)
        if count == 0:
            return
        self._delete(_hash, details)

    def _delete_many(self, _hash: str, data: List[Dict]):
        pass

    def delete_many(self, query: Dict[str, Any], data: List[Dict]):
        if not self.helpers.validate_query(query) or len(data) == 0:
            return

        updated_list = [self.connection.update_dict(query, x) for x in data]
        _hash = self.helpers.generate_hash(query)
        count = self.count(_hash)
        if count == 0:
            return
        self._delete_many(_hash, updated_list)

    def _delete_all(self, _hash: str):
        rlock = f"{_hash}:lock"
        relative_time_key = f"{_hash}:rlist"
        absolute_time_key = f"{_hash}:alist"
        with self.connection.lock(rlock):
            keys = self.connection.zrange(relative_time_key, 0, -1, withscores=True)

            values = [key[0] for key in keys]
            if len(values) == 0:
                return

            self.connection.zrem(relative_time_key, *values)
            self.connection.zrem(absolute_time_key, *values)

    def delete_all(self, query: Dict[str, Any]):
        if not self.helpers.validate_query(query):
            return

        _hash = self.helpers.generate_hash(query)
        self._delete_all(_hash)

    """
        Query commands
        ---
        All of the query information we need to operate on.


        The query functions we work on:

        1. `query` - Get all of the records related to a given key.
        2. `query_latest` - Get the `n` latest records according to our query parameters.
        3. `query_between` - Query between two epoch times.
        4. `query_before` - Get everything before an epoch time.
        5. `query_after` - Get everything after epoch time.    
    """

    def query_all(self, query: Dict[str, Any]):
        """ Same as query_all """
        if not self.helpers.validate_query(query):
            return []

        _hash = self.helpers.generate_hash(query)
        count = self.count(_hash)
        if count == 0:
            return []

        relative_time_key = f"{_hash}:rlist"
        absolute_time_key = f"{_hash}:alist"
        rlock = f"{_hash}:lock"
        with self.connection.pipeline() as pipe:
            pipe.watch(relative_time_key)
            pipe.watch(absolute_time_key)
            rkeys = pipe.zrange(relative_time_key, 0, -1, withscores=True)
            akeys = pipe.zrange(absolute_time_key, 0, -1, withscores=True)
            combined = self.helpers.combine_results(akeys, rkeys)
            pipe.execute()
        return combined

    def query_latest(self, _query: Dict[str, Any], abs_rel="absolute", limit: int = 10):
        if not self.helpers.validate_query(_query) or abs_rel not in [
            "absolute",
            "relative",
        ]:
            return {}
        _hash = self.helpers.generate_hash(_query)
        _current_key = self.helpers.dynamic_key(_hash, abs_rel)
        with self.connection.pipeline() as pipe:
            # with pipe.lock(rlock):
            pipe.watch(_current_key)
            count = self.count(_hash, pipe=pipe)
            if count == 0:
                return {}

            keys = pipe.zrange(_current_key, -1, -1, withscores=True)
            if len(keys) == 0:
                return {}
            pipe.execute()
        combined = self.helpers.combined_abs_rel(keys, abs_rel=abs_rel)
        return combined[-1]

    def query_latest_many(
        self, _query: Dict[str, Any], abs_rel="absolute", limit: int = 10
    ):
        if not self.helpers.validate_query(_query) or abs_rel not in [
            "absolute",
            "relative",
        ]:
            return {}
        _hash = self.helpers.generate_hash(_query)
        count = self.count(_hash)
        if count == 0:
            return {}

        _current_key = self.helpers.dynamic_key(_hash, abs_rel)
        keys = self.connection.zrange(_current_key, -limit, -1, withscores=True)
        if len(keys) == 0:
            return {}
        combined = self.helpers.combined_abs_rel(keys, abs_rel=abs_rel)
        return combined

    def query_between(
        self,
        _query: Dict[str, Any],
        min_epoch: float,
        max_epoch: float,
        abs_rel: str = "absolute",
    ):
        if not self.helpers.validate_query(_query) or abs_rel not in [
            "absolute",
            "relative",
        ]:
            return []
        _hash = self.helpers.generate_hash(_query)
        count = self.count(_hash)
        if count == 0:
            return []

        _current_key = self.helpers.dynamic_key(_hash, abs_rel)
        keys = self.connection.zrangebyscore(
            _current_key, min=min_epoch, max=max_epoch, withscores=True
        )
        combined = self.helpers.combined_abs_rel(keys, abs_rel=abs_rel)
        return combined

    def query_latest_by_time(
        self, _query: Dict[str, Any], max_epoch, abs_rel="absolute", limit: int = 10
    ):
        """ Get the closest item to a given timestamp. Simply pass in an epoch then watch things fly"""
        if not self.helpers.validate_query(_query) or abs_rel not in [
            "absolute",
            "relative",
        ]:
            return []
        _hash = self.helpers.generate_hash(_query)
        count = self.count(_hash)
        if count == 0:
            return []
        _current_key = self.helpers.dynamic_key(_hash, abs_rel)

        keys = self.connection.zrangebyscore(
            _current_key, min=max_epoch, max="+inf", start=0, num=1, withscores=True
        )
        combined = self.helpers.combined_abs_rel(keys, abs_rel)
        if len(combined) == 0:
            return {}
        return combined[0]

    def query_before(self, _query: Dict[str, Any], max_epoch, abs_rel="absolute"):
        if not self.helpers.validate_query(_query) or abs_rel not in [
            "absolute",
            "relative",
        ]:
            return []
        _hash = self.helpers.generate_hash(_query)
        count = self.count(_hash)
        if count == 0:
            return []
        _current_key = self.helpers.dynamic_key(_hash, abs_rel)
        keys = self.connection.zrangebyscore(
            _current_key, min="-inf", max=max_epoch, withscores=True
        )
        combined = self.helpers.combined_abs_rel(keys, abs_rel)
        return combined

    def query_after(self, _query: Dict[str, Any], min_epoch, abs_rel="absolute"):
        if not self.helpers.validate_query(_query) or abs_rel not in [
            "absolute",
            "relative",
        ]:
            return []
        _hash = self.helpers.generate_hash(_query)
        count = self.count(_hash)
        if count == 0:
            return []
        _current_key = self.helpers.dynamic_key(_hash, abs_rel)
        keys = self.connection.zrangebyscore(
            _current_key, min=min_epoch, max="+inf", withscores=True
        )
        combined = self.helpers.combined_abs_rel(keys, abs_rel)
        return combined

    """
        Other Functions
    """

    def _reset_count(self, query: Dict[str, Any], mongo_data: List[Dict]):
        """ Reset the count for the current mongodb query. We do this by adding records in mongo back into redis. """
        if len(mongo_data) == 0:
            return

        _hash = self.helpers.generate_hash(query)
        phindex = self.connection.incr("placeholder:index")
        delindex = self.connection.incr("deletion:index")
        _hash_key = f"{_hash}:list"
        _hash_placeholder = f"{_hash}:{phindex}"
        _hash_del = f"{_hash}:{delindex}"

        serialized_mongo = [orjson.dumps(mon) for mon in mongo_data]
        rlock = f"{_hash}:lock"
        with self.connection.pipeline() as pipe:
            with pipe.lock(rlock):
                pipe.watch(_hash)
                pipe.watch(_hash_placeholder)
                pipe.watch(_hash_key)
                pipe.connection.rpush(_hash_placeholder, *serialized_mongo)
                pipe.connection.rename(_hash_key, _hash_del)
                pipe.connection.rename(_hash_placeholder, _hash_key)

                pipe.execute()

        self._delete_all(_hash_del)

    def reset(self, query: Dict[str, Any], mongo_data: List[Dict]):
        if not self.helpers.validate_query(query):
            return
        self.pool.schedule(self._reset_count, args=(query, mongo_data))

    def count(self, _hash: str, pipe: Optional[Pipeline] = None) -> int:
        # ZCARD to get the length of the zset
        _count_hash = f"{_hash}:alist"
        if pipe is not None:
            pipe.watch(_count_hash)
            count = pipe.zcard(_count_hash)
        else:
            count = self.connection.zcard(_count_hash)
        return int(count)

    def max_score(self, _hash: str, pipe: Optional[Pipeline] = None):
        """max_score Get Max Score

            Get the max score given a key. We get the maximum score and cache it. 
            If we change any information we can swap the cache out dynamically to increase access speed.
        """
        _count_hash = f"{_hash}:alist"
        if pipe is not None:
            pipe.watch(_count_hash)
            count = pipe.zrangebyscore(min="+inf", max="-inf", start=0, num=1, withscores=True)
        else:
            count = self.connection.zrangebyscore(min="+inf", max="-inf", start=0, num=1, withscores=True)
        
        if count is None:
            return float(0.0)
        
        if isinstance(count, dict) and len(count) > 0:
            return float(count.keys()[0])
        return float(0.0)
    

    def min_score(self, _hash: str, pipe: Optional[Pipeline] = None):
        """min_score Get Minimum Score

            Get the min score given a key. We get the maximum score and cache it. 
            If we change any information we can swap the cache out dynamically to increase access speed.
        """
        _count_hash = f"{_hash}:alist"
        if pipe is not None:
            pipe.watch(_count_hash)
            count = pipe.zrangebyscore(min="-inf", max="+inf", start=0, num=1, withscores=True)
        else:
            count = self.connection.zrangebyscore(min="-inf", max="+inf", start=0, num=1, withscores=True)
        
        if count is None:
            return float(0.0)
        
        if isinstance(count, dict) and len(count) > 0:
            return float(count.keys()[0])
        return float(0.0)


import os

import time
import warnings
from copy import copy
from pprint import pprint
from typing import Any, Dict, List
warnings.simplefilter(action='ignore', category=FutureWarning)


from addict import Dict as ADict
from cerberus import Validator
from eliot import log_call, to_file
from loguru import logger
from redis.exceptions import ResponseError
from redisearch import Client, Query

from jamboree.utils.core import consistent_hash
from jamboree.utils.support.search import (InsertBuilder, QueryBuilder,
                                           is_gen_type, is_generic, is_geo,
                                           is_nested, is_queryable_dict,
                                           name_match, to_field, to_str)
from jamboree.utils.support.search.assistance import Keystore

from cytoolz import unique

logger.disable(__name__)
"""

    # NOTE

    Basic CRUD operations for the search handler. 
"""

REDIS_PORT = int(os.getenv("REDIS_PORT", "6379"))
REDIS_HOST = str(os.getenv("REDIS_HOST", "localhost"))

def split_doc(doc):
    return doc.id, ADict(**doc.__dict__)

def dictify(doc):
    item = ADict(**doc.__dict__)
    item.pop("super_id", None)
    item.pop("payload", None)
    return item

class BaseSearchHandlerSupport(object):
    def __init__(self):
        self._requirements_str = {
            
        }
        self._subkey_names = set()
        self._indexable = set()
        self.__indexable = []
        self._index_key:str = ""
        self._sub_fields = {}
        self.insert_builder = InsertBuilder()
        self.query_builder = QueryBuilder()
        self.keystore = Keystore()
        self.added = set()
        # Boolean explaining if this is a subquery
        self.is_sub_key = False
    
    
    @property
    def indexable(self):
        return self.__indexable
    
    @property
    def subnames(self):
        return self._subkey_names
    @property
    def index(self):
        """Index key for the requirements"""
        return self._index_key
    
    @index.setter
    def index(self, _index):
        """Index key for the requirements"""
        self._index_key = _index
    
    @property
    def subfields(self):
        return self._sub_fields
    
    def process_subfields(self):
        for key in self.subnames:
            self._sub_fields[key] = f"{self.index}:{key}"
    
    def process_requirements(self, _requirements:dict):
        """
            Process the required fields. That includes:
            
            1. Creating a requirements string. That's so we can create a key representing the field that exist.
            2. Listing all of the subkeys that we'd need to take in consideration.
            3. Creating an index hash to locate all relavent documents
            4. Creation of a list of fields so we can create a schema at that index hash
            5. Creation of all subkeys so we can quickly access them by name later
            
        """
        for k, v in _requirements.items():
            if is_generic(v):
                sval = to_str(v)
                _agg = f"{k}:{sval}"
                if _agg not in self.added:
                    self.added.add(_agg)
                    self._requirements_str[k] = sval
                    field = to_field(k, sval)
                    
                    self.__indexable.append(field)
                continue
                
            if v == dict:
                _agg = f"{k}:SUB"
                if _agg not in self.added:
                    self.added.add(_agg)
                    self._requirements_str[k] = "SUB"
                    self.subnames.add(k)
                continue

            if is_geo(v):
                _agg = f"{k}:GEO"
                if _agg not in self.added:
                    self.added.add(_agg)
                    self._requirements_str[k] = "GEO"
                    self.__indexable.append(to_field(k, "GEO"))

                continue
        
        # self._indexable = set(unique(self._indexable, key=lambda x: x.redis_args()[0]))
        if not self.is_sub_key:
            self._index_key = consistent_hash(self._requirements_str)
            self.process_subfields()

    def is_sub(self, name:str) -> bool:
        """ Check to see if this is a subfield """
        return name in self.subnames

    def is_queryable(self, _dict):
        if isinstance(_dict, dict):
            if is_queryable_dict(_dict):
                return True
        return False

    def is_valid_sub_key_information(self, subkey_dict:dict):
        """ Check to see if the subkey is valid"""
        
        if len(subkey_dict) == 0:
            return False
        
        
        # Run validation to see if all of the keys are reducible to a type and base type
        for k, v in subkey_dict.items():
            if is_generic(v):
                continue
            if isinstance(v, dict):
                if not is_queryable_dict(v):
                    logger.error(f"{k} is not valid")
                    return False
        return True

    def queryable_to_type(self, _dict:dict):
        """ Converts a queryable dictionary into a type"""
        dtype = _dict['type']
        if dtype == "GEO":
            return "GEO"
        elif dtype == "TEXT":
            return str
        elif dtype == "BOOL":
            return bool
        elif dtype == "NUMERIC":
            return float        
        elif dtype == "TAG":
            return list

    def loaded_dict_to_requirements(self, _dict:dict):
        """ 
            # Loaded Dict To Requirements
            
            Convert a dictionary into a requirements dict. 

            Use to create a requirements

            Returns an empty dict if nothing is there.
        """
        req = {}
        for k, v in _dict.items():
            _ktype = type(v)
            if is_generic(_ktype):
                req[k] = _ktype
            if self.is_queryable(v):
                req[k] = self.queryable_to_type(v)
                
        return req


    def reset_builders(self):
        self.insert_builder = InsertBuilder()
        self.query_builder = QueryBuilder()


import redis
from redis import Redis
from typing import Optional, Any
from jamboree.utils.helper import Helpers
from jamboree.storage.files.redisify import RedisFileProcessor, RedisFileConnection
from jamboree.base.processors.abstracts import FileProcessor


class JamboreeFileProcessor(FileProcessor):
    def __init__(self) -> None:
        self._redis:Optional[Redis] = None
        self._redis_conn = RedisFileConnection()
        self.helpers = Helpers()
    
    @property
    def rconn(self) -> redis.client.Redis:
        if self._redis is None:
            raise AttributeError("You've yet to add a redis connection")
        return self._redis
    
    @rconn.setter
    def rconn(self, _redis:redis.client.Redis):
        self._redis = _redis

    @property
    def redis_conn(self) -> RedisFileConnection:
        if self._redis_conn is None:
            raise AttributeError("Redis connection hasn't been set")
        return self._redis_conn
    
    @redis_conn.setter
    def redis_conn(self, _rconn: RedisFileConnection):
        self._redis_conn = _rconn

    def initialize(self):
        """ Initialize database connections. Use this so we can use the same connections for search, files, and events. """
        self.redis_conn = RedisFileConnection()
        self.redis_conn.conn = self.rconn

    def _validate_query(self, query: dict):
        """ Validates a query. Must have `type` and a second identifier at least"""
        if 'type' not in query:
            return False
        if not isinstance(query['type'], str):
            return False
        if len(query) < 2:
            return False
        return True
    

    """ 
        These are the basic functions. We'll create more functions to handle the different scenarios. 

        NOTE:
            * Eventually we'll add rocksdb and a check_local=False flag to explain that we want to search rocksdb first.
            * We can add a changed flag into redis. 
                * If we've changed a model on a different machine we can just say check_local=False
                * Otherwise we can set check_local=True and force retrieval from redis.
                * If we were to do that, we'd have to separate all of the setup functions inside of the redis handler
    """

    
    def save(self, query:dict, obj:Any, **kwargs):
        if not self._validate_query(query):
            raise ValueError("Query isn't valid")
        
        self.redis_conn.save(query, obj, **kwargs)
    
    def query(self, query:dict, **kwargs):
        if not self._validate_query(query):
            raise ValueError("Query isn't valid")
        data = self.redis_conn.query(query, **kwargs)
        return data

    def delete(self, query:dict, **kwargs):
        if not self._validate_query(query):
            raise ValueError("Query isn't valid")
        self.redis_conn.delete(query, **kwargs)

    def absolute_exists(self, query: dict, **kwargs):
        if not self._validate_query(query):
            raise ValueError("Query isn't valid")
        return self.redis_conn.absolute_exists(query, **kwargs)
    
    
    


from typing import List
class text(object):
    @staticmethod
    def exact(term):
        return {
            "type": "TEXT",
            "is_filter": True,
            "values": {
                "term": term,
                "is_exact": True
            }
        }
    
    @staticmethod
    def fuzzy(term):
        return {
            "type": "TEXT",
            "is_filter": True,
            "values": {
                "term": f"%{term}%",
                "is_exact": False
            }
        }
    
    @staticmethod
    def orlist(terms:List[str], is_bundle=False):
        _term = text.orliststr(terms, is_bundle)
        return {
            "type": "TEXT",
            "is_filter": True,
            "values": {
                "term": _term,
                "is_exact": False
            }
        }

    @staticmethod
    def orliststr(terms:List[str], is_bundle=False):
        if len(terms) == 0:
            return ""
        _term = "|".join(terms)
        if is_bundle:
            _temp = f"({_term})"
            _term = _temp
        return _term
    

class tags(object):
    
    @staticmethod
    def andfieldstr(field, items:List[str]):
        if len(items) == 0:
            return ""
        
        and_fields_str = [f"{field}:{item} " for item in items]
        return and_fields_str


from abc import ABC
from jamboree.utils.helper import Helpers
from pebble.pool import ThreadPool
from multiprocessing import cpu_count
from typing import Union, Optional
from redis import Redis
from redis.client import Pipeline


class DatabaseConnection(ABC):
    def __init__(self) -> None:
        self._connection: Optional[Union[Redis, Pipeline]] = None
        self.helpers = Helpers()
        self._pool = ThreadPool(max_workers=(cpu_count() * 2))

    @property
    def connection(self) -> Union[Redis, Pipeline]:
        if self._connection is None:
            raise AttributeError(
                "You haven't added a main database connection as of yet."
            )
        return self._connection

    @connection.setter
    def connection(self, _conn: Union[Redis, Pipeline]):
        self._connection = _conn

    @property
    def pool(self) -> ThreadPool:
        return self._pool

    @pool.setter
    def pool(self, _pool: ThreadPool):
        self.pool = _pool

    """ Save commands """

    def save(self, query):
        raise NotImplementedError("save not implemented")

    def save_many(self, query):
        raise NotImplementedError("save_many not implemented")

    """
        Update commands
    """

    def update_single(self, query):
        raise NotImplementedError("update_single not implemented")

    def update_many(self, query):
        raise NotImplementedError("update_many not implemented")

    """
        Delete Commands
    """

    def delete(self, query):
        raise NotImplementedError("delete function not implemented yet.")

    def delete_many(self, query):
        raise NotImplementedError("delete_many function not implemented yet.")

    def delete_all(self, query):
        raise NotImplementedError("delete_all not implemented")

    """ 
        Query commands
    """

    def query_latest(self):
        raise NotImplementedError("query_latest not implemented")

    def query_latest_many(self):
        raise NotImplementedError("query_latest_many not implemented")

    def query_between(self):
        raise NotImplementedError("query_between not implemented")

    def query_before(self):
        raise NotImplementedError("query_before not implemented")

    def query_after(self):
        raise NotImplementedError("query_after not implemented")

    def query_all(self):
        pass

    """ Other Functions """

    def reset(self, query):
        raise NotImplementedError("delete_all not implemented")

    def count(self):
        raise NotImplementedError("update_many not implemented")

    def general_lock(self, query: dict):
        raise NotImplementedError("general_lock not implemented")




"""
    A class that holds all of the helper functions.
"""
import base64
import uuid
from abc import ABC
from copy import copy
from typing import Any, Dict, List, Set

import maya
import orjson
import pandas as pd
import ujson


class Helpers(object):
    def __init__(self) -> None:
        pass

    def generate_hash(self, query: dict) -> str:
        _hash = ujson.dumps(query, sort_keys=True)
        _hash = base64.b64encode(str.encode(_hash))
        _hash = _hash.decode("utf-8")
        return _hash

    def hash_to_dict(self, _hash: str) -> str:
        # TODO: Convert Hash To Dict
        # NOTE: Not tested. Please make sure to test it
        __hash = base64.b64decode(_hash).decode("utf-8")
        _hash_json = ujson.loads(__hash)
        return _hash_json

    def add_time(self, item: dict, _time: float, rel_abs="absolute"):
        if rel_abs == "absolute":
            item["timestamp"] = _time
        else:
            item["time"] = _time
        return item

    def convert_to_storable_json(self, json_string: str):

        savable = {}
        for key, value in orjson.loads(json_string).items():
            item_json = orjson.dumps(value)
            timestamp = float(key) * 0.001
            savable[item_json] = timestamp
        return savable

    def convert_to_storable(self, items: list):
        savable = {}
        for item in items:
            timestamp = float(item.pop("timestamp", maya.now()._epoch))
            item_json = orjson.dumps(item)
            savable[item_json] = timestamp

        return savable

    def convert_to_storable_abs(self, items: list):
        savable = {}
        for item in items:
            timestamp = float(item.pop("timestamp", maya.now()._epoch))
            item_json = orjson.dumps(item)
            savable[item_json] = timestamp

        return savable

    def convert_to_storable_relative(self, items: list):
        savable = {}
        for item in items:
            timestamp = float(item.pop("time", maya.now()._epoch))
            item_json = orjson.dumps(item)
            savable[item_json] = timestamp

        return savable
    def dual_storable(self, items: list):
        relative = self.convert_to_storable_relative(items)
        absolute = self.convert_to_storable_abs(items)
        return {"relative": relative, "absolute": absolute}

    def convert_dataframe_to_storable_item(self, df: pd.DataFrame) -> dict:
        data = df.astype(str)
        data_json = data.to_json(orient="index")
        value = self.convert_to_storable_json(data_json)
        return value

    def get_current_abs_time(self, data: dict):
        _data = copy(data)
        for k, v in data.items():
            _data[k] = maya.now()._epoch
        return _data

    def generate_dicts(self, data: dict, _time: float, timestamp: float):
        relative = copy(data)
        absolute = copy(data)
        relative["time"] = _time
        absolute["timestamp"] = _time
        return {"relative": relative, "absolute": absolute}

    def is_abs_rel(self, query_type: str):
        if query_type in ["absolute", "relative"]:
            return True

        return False

    def is_zero_time(self, item: Dict[str, Any]) -> bool:
        """ 
            Check to see if any of the items has a time of zero:

            **Parameters:**
                - items: List items that represent a sequence of dictionaries from the data store.
            **Response:**
                - Is false if there are errors in the timestamp. Or if time or timestamp is 0
                - return Boolean
        """
        if "time" not in item and "timestamp" not in item:
            return False

        if "time" in item and "timestamp" in item:
            if item["time"] == 0 or item["timestamp"] == 0:
                return False
        if "time" in item:
            if item["time"] == 0:
                return False
        if "timestamp" in item:
            if item["timestamp"] == 0:
                return False
        return True

    def dictify(self, azset: List[Set], rzset: List[Set]):
        """Create a dictionary"""
        if len(azset) == 0 or len(rzset) == 0:
            return {}
        adict = {}
        for azs in azset:
            item, time = azs
            current_item = adict.get(item, {})
            current_item["timestamp"] = time
            adict[item] = current_item

        for rzs in rzset:
            item, time = rzs
            current_item = adict.get(item, {})
            current_item["time"] = time
            adict[item] = current_item

        return adict

    def add_event_id(self, event: dict):
        event["event_id"] = uuid.uuid4().hex
        return event

    def add_event_ids(self, data: List[Dict[str, Any]]):
        """ Add event ids to a list of events. Use for a save many query. """
        events = [self.add_event_id(x) for x in data]
        return events

    def check_time(
        self,
        _time: float = None,
        _timestamp: float = None,
        local_time: float = None,
        local_timestamp: float = None,
    ):
        current_time = maya.now()._epoch

        if local_time is not None:
            _time = local_time
        elif _time is None:
            _time = current_time
        if local_timestamp is not None:
            _timestamp = local_timestamp
        elif _timestamp is None:
            _timestamp = current_time

        return {"time": _time, "timestamp": _timestamp}

    def deserialize_dicts(self, dictified: dict):
        _deserialized = []
        for key, value in dictified.items():
            _key = orjson.loads(key)
            _key["time"] = value.get("time", maya.now()._epoch)
            _key["timestamp"] = value.get("timestamp", maya.now()._epoch)
            _deserialized.append(_key)
        return _deserialized

    def separate_time_data(
        self, data: dict, _time: float = None, _timestamp: float = None
    ):
        local_time = data.pop("time", None)
        local_timestamp = data.pop("timestamp", None)
        timing = self.check_time(_time, _timestamp, local_time, local_timestamp)
        return data, timing

    def validate_query(self, query: dict):
        """ Validates a query. Must have `type` and a second identifier at least"""
        if "type" not in query:
            return False
        if not isinstance(query["type"], str):
            return False
        if len(query) < 2:
            return False
        return True

    def update_dict(self, query: dict, data: dict):
        query = copy(query)
        timestamp = maya.now()._epoch
        query["timestamp"] = timestamp
        data.update(query)
        return data

    def update_dict_no_timestamp(self, query: dict, data: dict):
        query = copy(query)
        data = copy(data)
        data.update(query)
        data.pop("timestamp", None)
        return data

    def back_to_dict(self, list_of_serialized: list):
        deserialized = []
        if len(list_of_serialized) == 1:
            return orjson.loads(list_of_serialized[0])

        for i in list_of_serialized:
            deserialized.append(orjson.loads(i))
        return deserialized

    def search_one(self, item: dict, query: dict):
        all_bools = []
        for q in query:
            if q in item:
                if query[q] == query[q]:
                    all_bools.append(True)
                else:
                    all_bools.append(False)
            else:
                all_bools.append(False)
        return any(all_bools)

    def dynamic_key(self, _hash, absrel):
        """ Get the absolute key. """
        current_key = ""
        if absrel == "absolute":
            current_key = f"{_hash}:alist"
        else:
            current_key = f"{_hash}:rlist"
        return current_key

    def combine_results(self, akeys, rkeys):
        dicts = self.dictify(akeys, rkeys)
        dicts.pop(b'{"placeholder": "place"}', None)
        combined = self.deserialize_dicts(dicts)
        return combined

    def combined_abs_rel(self, keys, abs_rel="absolute"):
        blank_keys = [(b'{"placeholder": "place"}', 0)]
        dicts = []
        if abs_rel == "absolute":
            dicts = self.dictify(keys, blank_keys)
        else:

            dicts = self.dictify(blank_keys, keys)
        dicts.pop(b'{"placeholder": "place"}', None)
        combined = self.deserialize_dicts(dicts)
        return combined

    def convert_to_storable_json_list(self, json_string) -> list:
        converted_list = []
        for key, value in orjson.loads(json_string).items():
            item_json = value
            item_json["time"] = float(key) * 0.001
            converted_list.append(item_json)
        return converted_list

    def convert_to_storable_no_json(self, json_string: str) -> list:
        converted_list = []
        for key, value in orjson.loads(json_string).items():
            item_json = value
            item_json["time"] = float(key) * 0.001
            converted_list.append(item_json)
        return converted_list

    def convert_dataframe_to_storable_item_list(self, df: pd.DataFrame) -> list:
        data_json = df.to_json(orient="index")
        value = self.convert_to_storable_no_json(data_json)
        return value

    def standardize_record(self, record):
        closing_record = copy(record)
        if "Close" in record:
            closing_record.pop("Close", None)
            closing_record["close"] = record["Close"]
        if "Open" in record:
            closing_record.pop("Open", None)
            closing_record["open"] = record["Open"]
        if "Low" in record:
            closing_record.pop("Low", None)
            closing_record["low"] = record["Low"]
        if "High" in record:
            closing_record.pop("High", None)
            closing_record["high"] = record["High"]
        if "Volume" in record:
            closing_record.pop("Volume", None)
            closing_record["volume"] = record["Volume"]
        if "Adj Close" in record:
            closing_record.pop("Adj Close", None)
            closing_record["adj_close"] = record["Adj Close"]
        return closing_record

    def standardize_outputs(self, records: List[Dict[str, Any]]):
        if len(records) == 0:
            return []
        _records = [self.generic_standardize(rec) for rec in records]
        return _records

    def generic_standardize(self, record: dict):
        if bool(record) == False:
            return {}

        _record = {}
        for k, v in record.items():
            # Normalize
            k = str(k)
            k = k.translate({ord(c): "" for c in "!@#$%^&*()[]{};:,./<>?|`~-=_+"})
            k = k.lower()
            k = k.replace(" ", "_")
            _record[k] = v
        return _record



from jamboree.handlers.processors import DataProcessorsAbstract
import pandas as pd
import numpy as np
import scipy.stats as stats


class DynamicResample(DataProcessorsAbstract):
    def __init__(self, name, **kwargs) -> None:
        self.time_info = {
            "years": 0,
            "months": 0,
            "weeks":0,
            "days": 0,
            "hours": 0,
            "minutes": 0,
            "seconds": 0
        }
        self.time_formatting = {
            "seconds": "S",
            "minutes": "T",
            "hours": "H",
            "days": "D",
            "weeks":"W",
            "months": "M",
            "years": "Y"
        }

        self.base = 0
        super().__init__(name, **kwargs)


    def set_settings(self, **kwargs):
        """ Updates the time information"""
        self.time_info['seconds'] = kwargs.get("seconds",self.time_info['seconds'])
        self.time_info['minutes'] = kwargs.get("minutes",self.time_info['minutes'])
        self.time_info['hours'] = kwargs.get("hours",self.time_info['hours'])
        self.time_info['days'] = kwargs.get("days",self.time_info['days'])
        self.time_info['weeks'] = kwargs.get("weeks",self.time_info['weeks'])
        self.time_info['months'] = kwargs.get("months",self.time_info['months'])
        self.time_info['years'] = kwargs.get("years",self.time_info['years'])
        self.base = kwargs.get("base", self.base)

    def validate_existing_times(self):
        checkable_list = self.time_info.values()
        all_zero = all(x==0 for x in checkable_list)
        if all_zero:
            self.time_info['hours'] = 1

    def generate_time_string(self):
        self.validate_existing_times()
        final_string = ""
        for name, time_amount in self.time_info.items():
            if time_amount == 0:
                continue
            elif time_amount == 1:
                final_string = final_string + self.time_formatting.get(name)
                continue
            
            time_format = self.time_formatting.get(name)
            final_string = final_string + f"{time_amount}{time_format}"
        
        return final_string

    def process(self, data: pd.DataFrame) -> pd.DataFrame:
        if not (isinstance(data, pd.DataFrame)):
            return pd.DataFrame()
        
        if data.empty:
            return pd.DataFrame()
        # Do preprocessing here
        
        dtypes = data.dtypes
        aggregate_command = {}
        for k, v in dtypes.items():
            if np.issubdtype(v, np.number):
                command = {k:'mean'}
                aggregate_command.update(command)
                continue
            else:
                command = {k: lambda x: stats.mode(x)[0]}
                aggregate_command.update(command)
                continue
        
        rule = self.generate_time_string()

        if self.base == 0:
            resampled = data.resample(rule).apply(aggregate_command)
            return resampled
        resampled = data.resample(rule, base=self.base).apply(aggregate_command)
        return resampled

if __name__ == "__main__":
    import pandas_datareader.data as web
    data_msft = web.DataReader('MSFT','yahoo',start='2008/1/1',end='2020/3/8').round(2)
    mrsample = DynamicResample("modin", days=7)

    remicro = mrsample.process(data_msft)
    print(remicro)
    # time_str = mrsample.generate_time_string()
    # print(time_str)


from abc import ABC
from typing import List

class FileProcessor(ABC):
    """ 
        # File processor abstract. 
        
        Deals with all blobs and files..
    """
    
    def initialize(self):
        pass

    def save(self, query: dict, data: dict, **kwargs):
        """ Save a single blob of data. """
        raise NotImplementedError

    def save_version(self, query, **kwargs):
        """ Save a single blob of data at a given version. """
        raise NotADirectoryError

    def query(self, query, **kwargs):
        """ Query a blob of data. Get the latest """
        raise NotImplementedError

    def query_version(self, query, **kwargs):
        """ Save an explicit version of data """
        raise NotImplementedError

    def delete(self, query, **kwargs):
        """ Delete the latest version of data """
        raise NotImplementedError
    
    def delete_version(self, query:dict, **kwargs):
        """ Delete a given version of data if it exist """
        raise NotImplementedError

    def delete_all(self, query:dict, **kwargs):
        """ Purge everything """
        raise NotImplementedError
    
    def absolute_exists(self, query:dict, **kwargs):
        raise NotImplementedError


from abc import ABC
from typing import Any


class DataProcessorsAbstract(ABC):
    """ DataProcessor is used to transform"""
    def __init__(self, name, **kwargs):
        self._name = name
        self.set_settings(**kwargs)

    def set_settings(self, **kwargs):
        raise NotImplementedError(
            "Need to set the settings you're expecting for this preprocessor"
        )
    
    def process(self, data:Any) -> Any:
        raise NotImplementedError(
            "A command to preprocess information and return that info."
        )




from abc import ABC
from typing import List


class LegacyProcessor(ABC):
    """ Use to make the main jamboree object still function without a problem. We have a few ways to explore this concept."""
    def save(self, query: dict, data: dict, abs_rel="absolute"):
        raise NotImplementedError


    def save_many(self, query: dict, data: List[dict], abs_rel="absolute"):
        raise NotImplementedError    


    def get_latest(self, query, abs_rel="absolute") -> dict:
        raise NotImplementedError


    def get_latest_many(self, query, abs_rel="absolute", limit=1000):
        raise NotImplementedError


    def get_between(self, query:dict, min_epoch:float, max_epoch:float, abs_rel:str="absolute") -> list:
        raise NotImplementedError


    def get_latest_by(self, query:dict, max_epoch, abs_rel="absolute", limit:int=10) -> dict:
        raise NotImplementedError


    def count(self, query: dict) -> int:
        raise NotImplementedError


    def remove_first(self, query: dict):
        raise NotImplementedError

    
    def pop_multiple(self, query: dict, limit: int):
        raise NotImplementedError


    def _bulk_save(self, query: dict, data: list):
        raise NotImplementedError


    def single_get(self, query:dict):
        raise NotImplementedError


    def single_set(self, query:dict, data:dict):
        raise NotImplementedError


    def single_delete(self, query:dict):
        raise NotImplementedError


"""
    Basic storage handler
    ---

"""

import copy
from typing import Any, Dict, Optional

import ujson
from addict import Dict as ADict

from jamboree import JamboreeNew
from jamboree.base.processors.abstracts import EventProcessor, Processor
from jamboree.handlers.base import BaseFileHandler, BaseHandler
from jamboree.utils.helper import Helpers

class BlobStorageHandler(BaseHandler):
    """ 
        A simple event store using a variation of databases.
        ---
        
        Currently uses zadd to work
    """

    def __init__(self):
        # print("DBHandler")
        self._entity = ""
        self._meta_type = "storage"
        self._required = {}
        self._query = {}
        self._data = {}
        self._is_event = True
        self._processor: Optional[Processor] = None
        self.event_proc: Optional[EventProcessor] = None
        self.main_helper = Helpers()
        self.changed_since_command = False

    def __setitem__(self, key, value):
        if bool(self.required):
            
            if key in self.required:
                self._query[key] = value
                return self._query
        self._data[key] = value
        self.changed_since_command = True
        return self._data

    def __getitem__(self, key):
        if key in self._query.keys():
            return self._query.get(key, None)
        else:
            if key in self._data.keys():
                return self._data.get(key, None)
        return None
    
    def setup_query(self, alt={}):
        query = copy.copy(self._query)
        query['type'] = self.entity
        query['mtype'] = self._meta_type
        query.update(alt)
        query.update(self._data)
        return query


    @property
    def is_event(self) -> bool:
        """ Determines if we're going to add event ids to what we're doing. We can essentially set certain conditions"""
        return self._is_event
    
    @is_event.setter
    def is_event(self, is_true:bool=False):
        self._is_event = is_true

    @property
    def processor(self) -> 'Processor':
        if self._processor is None:
            raise AttributeError("The Processor is missing")
        return self._processor
    
    @processor.setter
    def processor(self, _processor: 'Processor'):
        self._processor = _processor

    
    def clear_event(self) -> None:
        self._processor = None

    @property
    def entity(self):
        return self._entity

    @entity.setter
    def entity(self, _entity: str):
        self._entity = str(_entity)

    @property
    def required(self):
        return self._required

    @required.setter
    def required(self, _required: Dict[str, Any]):
        # check to make sure it's not empty
        self._required = _required

    @property
    def query(self):
        return self._query

    @query.setter
    def query(self, _query: Dict[str, Any]):
        if len(_query.keys()) > 0:
            self._query = _query

    def check(self):
        
        # self.processor
        # if self.event_proc is None:
        #     raise AttributeError("Event processor isn't available.")

        if (not bool(self._entity)) or (not bool(self._required)) or (not bool(self._query)):
            raise AttributeError(f"One of the key variables is missing.")

        for req in self._required.keys():
            _type = self._required[req]
            if req not in self._query:
                raise AttributeError(f"{req} is not in the requirements")
            if not isinstance(self._query[req], _type):
                raise AttributeError(f"{req} is not a {_type}")
        return True
    

    def save(self, data: dict, alt={}, is_overwrite=False):
        self.check()
        query = self.setup_query(alt)
        # Put settings here
        current_settings = ADict()
        current_settings.overwrite = is_overwrite
        self.processor.storage.save(query, data, **current_settings.to_dict())
        self.changed_since_command = False
    

    def save_version(self, data: dict, version:str, alt={}, is_overwrite=False):
        self.check()
        query = self.setup_query(alt)
        # Put settings here
        current_settings = ADict()
        self.processor.storage.save(query, data, **current_settings.to_dict())
        
        self.changed_since_command = False
    
    def absolute_exists(self, alt={}):
        self.check()
        query = self.setup_query(alt)
        # Put settings here
        current_settings = ADict()
        current_settings.is_force = self.changed_since_command
        avs = self.processor.storage.absolute_exists(query, **current_settings.to_dict())
        self.changed_since_command = False
        return avs

    def last(self, alt={}):
        self.check()
        query = self.setup_query(alt)
        current_settings = ADict()
        self.changed_since_command = False
        obj = self.processor.storage.query(query, **current_settings.to_dict())
        return obj
    
    def by_version(self, version:str, alt={}):
        """ Get the data by version. """
        self.check()
        query = self.setup_query(alt)
        current_settings = ADict()
        self.processor.storage.query(query, **current_settings.to_dict())
        self.changed_since_command = False
    
    def delete(self, query:dict, alt={}):
        self.check()
        query = self.setup_query(alt)
        current_settings = ADict()

        self.processor.storage.delete(query, **current_settings)
        self.changed_since_command = False
    
    def lock(self, alt={}):
        self.check()
        query = self.setup_query(alt)
        self.changed_since_command = False
        return self.processor.event.lock(query)
    
    def clear(self):
        """ Clear in-memory cache. To use with existence checks and rockdb """
        self.changed_since_command = True



import os
import re
import json
import time
import warnings
from contextlib import suppress
from copy import copy
from pprint import pprint
from typing import Any, Dict, List, Optional
warnings.simplefilter(action='ignore', category=FutureWarning)
# from cytoolz import unique

"""
NOTE: 

After working with this. I realized how much I fucked up. Will need to fix this. 
Not only did I not know what I was doing, I rushed it and didn't design correctly.

1. Sharpen the old interfaces (solid variables, consistent abstract classes)
2. separate code way more
3. Make take existing use cases (after I've finished this sprint) and make it easier
4. Reduce computation requirements with dicts (objects as often as possible)
5. Make the code super traceable
    - This means keeping track of alters that happen.
    - Maybe have a transition field we write to upon changing tables
    - When you create a schema log the means in which you created that schema. 
    - If you created a schema from a json, take note of that.
6. Make it more decoupled and composable
    - Take their styling, and figure out how to give it some magic. 
    - Use modifiers 
7. Figure out a better way of dealing with schemas (hashes all the way)
8. Move away all of the helper functions to different spots
9. Using classes as parameters to increase certainty in spots
10. Ways of tracking changes with hashes
11. Pipes in many of the places to reduce lookups
12. Have edge cases covered all the way (you'll figure that out through use)
13. Facade pattern
14. Practice use cases using a facade pattern first
    - Access the client directly if you need to.
    - Have controlled ways of messing with pipes
    - Move hella more code from event sourcing to this
    - Come up with automated adds to redisearch if possible. 
        - This will allow you to create dynamic schemas 
15. Come up with better relational patterns (before attempting graphs)
16. Lots of method objects if possible.
17. Do something entirely different with the query builder (use their interface more) and perhaps a builder pattern or something.
18. Reduce the number of loops, especially when not in C.
19. Have some native pagination done.
    - Configuration as a whole should be better. You improved that for managing the backtest, you can do it again for this.
20. Enums, More Constraints and Tests
21. Maybe having some predefined query wrappers would save time fucking with that
22. Do more efficient queries using the info()
23. Prevent fall off conditions for pipes
    - Wrap the copy the current processor when using a wrapped call.
24. Create a SchemaClass of your own
    `
        class RequirementsInteraction:
            # Used to track metadata fields
            description_fields = *fields 
            # Figure out a way to attach a psha to the requirements, even with different variables
            # Find a 
            field = TypeObject(**parameters) # parameters here describe how the data will be used
            field = TypeObject(**parameters)
            field = TypeObject(**parameters)
            field = TypeObject(**parameters)
            field = TypeObject(**parameters)
        
        def alternative(self, schema:SearchSchema):
            # Would search create two keys, first would be the default key, second would be the expanded one
        
        def prepare(self, redisearch_object):
            # if there was any alter command lately, we would
            pass
    `


"""

from addict import Dict as ADict
from cerberus import Validator
from loguru import logger
from redis.exceptions import ResponseError
from redisearch import Client, Query

from jamboree import Jamboree
from jamboree.base.processors.abstracts import Processor
from jamboree.utils.core import consistent_hash
from jamboree.utils.support.search import (BaseSearchHandlerSupport,
                                           InsertBuilder, QueryBuilder,
                                           is_generic, is_geo, is_nested,
                                           is_queryable_dict, name_match,
                                           to_field, to_str)



logger.disable(__name__)
"""

    # NOTE

    Basic CRUD operations for the search handler. 
"""


def split_doc(doc):
    return doc.id, ADict(**doc.__dict__)

def dictify(doc, is_id=True):
    item = ADict(**doc.__dict__)
    item.pop("super_id", None)
    item.pop("payload", None)
    if is_id == False:
        item.pop("id", None)

    return item

def single_doc_check_convert(doc):
    item = doc.__dict__
    item_conv = ADict(**item)
    item_conv_id = item_conv.pop("id", None)
    item_conv.pop("payload", None)
    
    return item_conv, bool(item_conv) 

def doc_convert(doc):
    item = doc.__dict__
    item_conv = ADict(**item)
    # item_conv_id = item_conv.pop("id", None)
    item_conv.pop("payload", None)
    return item_conv

class BaseSearchHandler(BaseSearchHandlerSupport):
    def __init__(self):
        super().__init__()
        self._entity = None
        # This will only be here as an example
        self.is_replacement = False
        self.current_replacement = None
        self.is_set_entity = False
                
        # Subs are all of the subfields we would need to search through
        self.subs:Dict[str, BaseSearchHandler] = {}


        self.current_doc_id = None
        self.current_doc_id_list = set()
        self.current_client = None
        
        self.print_sub = False
        self.use_sub_query = False

        self._super_ids = []
        self._sub_ids = []
        self.finished_alter = False



        self.search_sub = False
        self._processor:Optional[Processor] = None


    def __setitem__(self, key:str, value:Any):
        if key not in self.requirements.keys() and (not self.is_replacement):
            return
        self.is_set_entity = False
        self.current_client = None
        if isinstance(value, dict):
            if len(value) == 0: return
            self.handle_input_dict_key(key, value) 
        else:
            _instance_type = type(value)
            # check that the value is the right type
            if is_generic(_instance_type):
                _str_type = to_str(_instance_type)
                self.query_builder.insert_by_type_str(_str_type, key, value)
                self.insert_builder.insert_by_type_str(_str_type, key, value)


    @property
    def replacement(self):
        if self.current_replacement is None:
            self.current_replacement =  BaseSearchHandler()
            self.current_replacement.is_replacement = True
            self.current_replacement.insert_builder.is_replacement = True
            # self.current_replacement.
        return self.current_replacement

    @property
    def entity(self):
        if self._entity is None:
            raise AttributeError("You haven't set entity yet")
        return self._entity
    
    @entity.setter
    def entity(self, _entity:str):
        self._entity = _entity
    

    @property
    def processor(self):
        if self._processor is None:
            raise AttributeError("The processor hasn't been set yet.")
        return self._processor

    @processor.setter
    def processor(self, _processor:Processor):
        self._processor = _processor
        self.set_sub_processors()

    @property
    def requirements(self):
        return self._requirements_str
    
    @requirements.setter
    def requirements(self, _requirements:dict):
        """If we set it here we'd go through each dict item and create string version of each key"""
        # Document id will allow us to figure out which documents are involved with subkeys
        
        _requirements['entity'] = str
        _requirements['super_id'] = str
        self.process_requirements(_requirements)
        if not self.is_sub_key:
            self.create_sub_handlers()

    @property
    def dreq(self):
        return self._dreq
    
    @dreq.setter
    def dreq(self, _req):
        self._dreq = _req
        self.reset()
        self.requirements = _req
        self.replacement.requirements = _req

    @property
    def allrequirements(self):
        return self._dreq
    
    @allrequirements.setter
    def allrequirements(self, _req):
        self._dreq = _req
        self.reset()
        self.requirements = _req
        self.replacement.requirements = _req

    @property
    def doc_id(self):
        """ We get the current doc_id if it exists"""

        return self.current_doc_id

    @doc_id.setter
    def doc_id(self, _doc_id:str):
        self.current_doc_id = _doc_id


    @property
    def verbatim(self):
        return self.query_builder.build_exact()

    @property
    @logger.catch(ResponseError)
    def client(self):
        """client

        Get the client for the user. If it doesn't exist yet, create a new one with the given stop words. 
        
        For subkeys it adds fields if we've created them recently. 

        .. code-block:: python

            >>> self.client.add_document(_id, payload, **records)


        Returns
        -------
        [type]
            A redis connected client. Gets connection from Jamboree processor.
        """
        # self.processor
        if self.current_client is None:
            # We would insert a connection here. Use the connection from the search processor to operate.
            with suppress(ResponseError):
                self.current_client = Client(self.index, conn=self.processor.rconn)
                # print(self.indexable)
                if len(self.indexable) > 0:
                    self.current_client.create_index(self.indexable, stopwords=["but", "there", "these", "they", "this", "to"])
        
        if self.is_sub_key:
            if not self.finished_alter:
                for i in self.indexable:
                    with suppress(ResponseError):
                        self.current_client.alter_schema_add([i])
                self.finished_alter = True


        return self.current_client

    @property
    def general(self):
        return self.query_builder.general

    @general.setter
    def general(self, term:str):
        """ Push a general term into the query. It can only be done once. Don't put it to a filter key."""
        if not isinstance(term, str):
            logger.error("Term isn't a string")
            return
        self.query_builder.general = term
    
    

    """
        This is when things get weird
    """


    def create_sub_handlers(self):
        """ Creates subhandlers for the given index"""
        for name, subkey in self.subfields.items():
            subhandler = BaseSearchHandler()
            subhandler.is_sub_key = True
            subhandler.index = subkey
            subhandler.insert_builder.is_sub = True
            
            self.replacement.subs[name] = copy(subhandler)
            self.subs[name] = subhandler

    def set_sub_processors(self):
        """ If there are any sub queries, set processors to them """
        if len(self.subfields) > 0:
            self.use_sub_query = True
            for name in self.subfields.keys():
                self.subs[name].processor = self.processor
                with suppress(Exception):
                    self.replacement.subs[name].processor = self.processor

    def set_entity(self):
        if self.is_set_entity is False:
            self['entity'] = {
                "type": "TEXT",
                "is_filter": True,
                "values": {
                    "is_exact": True,
                    "term": self.entity
                }
            }
            self.is_set_entity = True



    def verbatim_docs(self):
        built = self.query_builder.build_exact()
        q = (
                Query(built)
                .no_stopwords()
                .paging(0, 1000000)
            )
        results = self.client.search(q)
        result_docs = results.docs
        return result_docs
    

    def general_docs(self):
        built = self.query_builder.build()
        q = (
                Query(built)
                .paging(0, 1000000)
            )
        results = self.client.search(q)
        result_docs = results.docs
        return result_docs

    
    def verbatim_sub_ids(self):
        super_id_set = set()
        sub_id_set = set()
        
        for key, sub in self.subs.items():
            sub.print_sub = True

            built = sub.query_builder.build()
            # logger.warning(built)
            built = built.strip(' ')
            is_falsy = (not built)
            if is_falsy:
                continue
            # logger.error(built)
            verb_items = sub.general_docs()
            current_super_ids = []
            current_subs = []
            for verb in verb_items:
                try:
                    _verb_id = verb.id
                    _super_id = verb.super_id
                    full_dict = verb.__dict__

                    self.keystore.add(_super_id, key, full_dict)
                    current_subs.append(_verb_id)
                    current_super_ids.append(_super_id)
                except Exception as e:
                    logger.error(str(e))
            
            if len(current_super_ids) > 0:
                if len(super_id_set) == 0:
                    super_id_set.update(current_super_ids)
                else:
                    super_id_set = super_id_set.intersection(current_super_ids)
            
            sub_id_set.update(current_subs)
        
        return list(super_id_set), list(sub_id_set)

    def verbatim_doc_ids(self):
        q = Query(self.verbatim).no_content().paging(0, 1000000)
        results = self.client.search(q)
        ids = [res.id for res in results.docs]
        return ids

    def handle_input_dict_key(self, name:str, item:dict):
        """ Figures out where to put the input dictionary for the query """
        if self.is_sub(name):
            # If this is a subkey we'll run the same operation again
            # Check to see if the subkey is empty and has information that is reducible to "type"
            self.use_sub_query = True
            self.search_sub = True
            reqs = self.loaded_dict_to_requirements(item)
            # logger.debug(reqs)
            self.subs[name].requirements = reqs
            for k, v in item.items():
                self.subs[name][k] = v
        else:
            # If it's not queryable don't try adding anything
            if not is_queryable_dict(item):
                return
            self.insert_builder.from_dict(name, item)
            self.query_builder.from_dict(name, item)


    def normal_find(self, limit_ids=None):
        built = self.query_builder.build()
        q = Query(built).paging(0, 1000000)
        # print(built)
        # print(q.query_string())
        if limit_ids is not None and len(limit_ids) > 0:
            q.limit_ids(*limit_ids)

        results = self.client.search(q)
        result_docs = results.docs
        return result_docs
    
    def normal_find_ids(self, limit_ids=None):
        _query = self.query_builder.build()
        q = Query(_query).no_content().paging(0, 1000000)
        if limit_ids is not None and len(limit_ids) > 0:
            q.limit_ids(*limit_ids)
        results = self.client.search(q)
        result_docs = results.docs
        return [res.id for res in result_docs] 



    def sub_find(self):
        sup_ids, sub_ids = self.verbatim_sub_ids()
        logger.success((sup_ids, sub_ids))
        if len(sub_ids) == 0: return []
        results = self.normal_find(limit_ids=sup_ids)
        results_dicts = []
        for result in results:
            _id, idict = split_doc(result)
            
            idict.pop("payload", None)
            subitems = self.keystore.get(_id)
            idict.update(subitems)
            results_dicts.append(idict)
        return results_dicts

    def normal_insert(self, allow_duplicates=False):
        if allow_duplicates == False:
            verbatim_docs = self.verbatim_docs()

            if len(verbatim_docs) > 0 and allow_duplicates == False:

                # Not adding docs because we're not allowing duplicates
                return verbatim_docs[0].id, False
        insert_variables = self.insert_builder.build()
        _doc_id = self.insert_builder.doc_id
        index_name = self.client.index_name
        fields = [i.redis_args()[0] for i in self.indexable]
        with logger.catch(message=f"{index_name} - {fields}", reraise=True):
            self.client.add_document(_doc_id, payload=_doc_id, **insert_variables)

        return _doc_id, True

    def sub_insert(self, allow_duplicates=False):
        _super_id, _did_insert = self.normal_insert(allow_duplicates=allow_duplicates)
        # logger.info(f'Did insert: {_did_insert}')
        if _did_insert:
            for key, sub in self.subs.items():
                if len(sub.insert_builder._insert_dict) > 0:
                    sub.insert_builder.super_id = _super_id
                    sub.normal_insert(allow_duplicates=True)
        return _super_id

    def find_sub_dictionaries(self, super_id):
        """ Finds a subdictionary by superid inside of the database. """
        # Should use the find within function for every subkey
        mega_dict = ADict()
        for key, sub in self.subs.items():
            key_dict = ADict()
            try:
                res = sub.client.search(f'"{super_id}"')
                if res.total == 0:
                    continue
                dd = [dictify(doc, False) for doc in res.docs]
                key_dict[key] = dd[0]
            except ResponseError:
                pass
            mega_dict.update(key_dict)
        return mega_dict

    
    def find(self):
        """Given the items we've set, find all matching items"""
        
        self.set_entity()
        self.keystore.reset()
        if self.use_sub_query and self.search_sub: 
            return self.sub_find()
        normal = self.normal_find()
        if len(self.subs) == 0:
            if len(normal) > 0:
                return [doc_convert(x) for x in normal]
            return normal
        ndicts = []
        for i in normal:
            # print(i)
            _i = dictify(i)
            mega = self.find_sub_dictionaries(_i.id)
            if len(mega) > 0:
                _i.update(mega.to_dict())
            ndicts.append(_i)
        return ndicts
    
    def pick(self, _id:str):
        """ 
            Given an id find the element with the top level id. We aren't searching lower level_ids. 
            
            After we pull all of the 
        """
        self.set_entity()
        self.keystore.reset()
        doc = self.client.load_document(_id)
        dd = doc.__dict__
        doc = ADict(**dd)
        _id = doc.pop("id", None)
        doc.pop("payload", None)
        doc_z = (len(doc) > 0)
        if len(self.subs) == 0:
            if not doc_z:
                return None
            doc.update({"id": _id})
            return doc
        
        if doc_z:
            sub_dicts = self.find_sub_dictionaries(_id)
            # if len(sub_dicts) > 0:
            doc.update(sub_dicts)
            doc.update({"id": _id})
            return doc
        
        return None


    def update(self):
        """
            # UPDATE

            Given the items or ID we've set, partial update every matching document. 
            If we have the document_ids already, replace those items
        """
        self.set_entity()
        self.keystore.reset()

        replacement_variables = self.replacement.insert_builder.build()
        if self.use_sub_query == False:
            doc_ids = self.verbatim_doc_ids()
            batcher = self.client.batch_indexer(chunk_size=len(doc_ids))
            for doc_id in doc_ids:
                batcher.add_document(doc_id, replace=True, partial=True, **replacement_variables)
            batcher.commit()
        else:
            sup_ids, sub_ids = self.verbatim_sub_ids()
            norm_ids = self.normal_find_ids(limit_ids=sup_ids)
            batcher = self.client.batch_indexer(chunk_size=len(norm_ids))
            for doc_id in norm_ids:
                batcher.add_document(doc_id, replace=True, partial=True, **replacement_variables)
            batcher.commit()


            for sub in self.subs.values():
                subreplacement = sub.insert_builder.build()
                if len(subreplacement) > 0:
                    subbatcher = sub.client.batch_indexer(chunk_size=len(sub_ids))
                    for _id in sub_ids:
                        self.client.add_document(_id, replace=True, partial=True, **subreplacement)
                    subbatcher.commit()
    
    def update_id(self, _id):
        self.set_entity()
        self.keystore.reset()
        doc = self.client.load_document(_id)
        doc_dict, is_exist = single_doc_check_convert(doc)
        
        if not is_exist:
            return
        
        replacement_variables = self.replacement.insert_builder.build()
        self.client.add_document(_id, replace=True, partial=True, **replacement_variables)
        doc = self.client.load_document(_id)
        # print(doc)
        # if len(self.subs) > 0:
        #     subreplacement = sub.insert_builder.build()
        #     print(subreplacement)

    # def insert_many(self, list_of_items):
    #     self.client.ba
    #     pass
    
    

    def insert(self, allow_duplicates=False):
        """
            # INSERT

            Given all of the items we've set, add documents
        """
        self.set_entity()
        self.keystore.reset()
        previous_id = None
        if self.use_sub_query:
            previous_id = self.sub_insert(allow_duplicates=allow_duplicates)
        else:
            previous_id, is_add = self.normal_insert(allow_duplicates=allow_duplicates)
        return previous_id



    def remove(self):
        """Remove all documents that match a query.

        Given a query, remove every document that matches the results of that query. 

        ::
            >>> search['name'] = 'sample_name'
            >>> search['category'] = 'sample_query'
            >>> search.remove() 


        """
        self.set_entity()
        self.keystore.reset()
        if self.use_sub_query and self.search_sub:
            removable = set()
            sup_ids, sub_ids = self.verbatim_sub_ids()
            norm_ids = self.normal_find_ids(limit_ids=sup_ids)
            removable = removable.intersection(sup_ids)
            removable = removable.intersection(norm_ids)


            [self.client.delete_document(_id) for _id in removable]
            for sub in self.subs.values():
                for _id in sub_ids:
                    sub.client.delete_document(_id)
        else:
            norm_ids = self.normal_find_ids()
            # print(norm_ids)
            [self.client.delete_document(_id) for _id in norm_ids]
    

    def reset(self):
        """Reset all local variables"""
        self.reset_builders()
        self.is_set_entity = True
        self.is_replacement = False
        self.current_replacement = None
        self.current_client = None
        self.use_sub_query = False

class ExampleSearchHandler(BaseSearchHandler):
    def __init__(self):
        super().__init__()
        self.entity = "example"
        self.dreq = {
            "name": str,
            "category": str,
            "subcategories": dict,
            "secondsub": dict,
            "live": bool,
            "loc": "GEO"
        }
        

def main():
    logger.enable(__name__)
    processor = Jamboree()
    example_handler = ExampleSearchHandler()
    example_handler.processor = processor
    example_handler['name'] = "Boi Gurl"
    example_handler['category'] = "marketsx"
    example_handler['sample_tags'] = ["four", "five", "six"]
    # TODO: Figure out subcategory update
    example_handler['subcategories'] = {
        "hello": "world",
        "country": "US",
    }
    example_handler['secondsub'] = {
        "my": "jac"
    }
    example_handler.replacement['secondsub'] = {
        "my": "jack"
    }
    example_handler['live'] = False

    # example_handler.remove()
    example_handler.insert(allow_duplicates=True)
    records = example_handler.find()
    print("\n\n")
    logger.warning((records, len(records)))

    records = example_handler.find()
    print("\n\n")
    logger.warning((records, len(records)))

    records = example_handler.find()
    print("\n\n")
    logger.warning((records, len(records)))

    # example_handler.update()
    # records = example_handler.find()
    # logger.success(records)




if __name__ == "__main__":
    main()



import re
import uuid
from typing import Dict
from loguru import logger
from jamboree.utils.support.search import (
    is_valid_geo, is_valid_bool, is_valid_numeric, is_valid_tags, is_valid_text
)
from jamboree.utils.support.search.assistance import inserter

class QueryBuilder(object):
    """
        # QueryBuilder

        Builds a redis query given a set of builder parameters.

        Some assumptions:
            - All tag queries
    """

    def __init__(self):
        self._query_set = {
            
        }
        self._boolean_fields = set()
        self._number_fields = set()
        self._text_fields = set()
        self._tag_fields = set()
        self._geo_fields = set()

        self.geo_query = {

        }
        self._general = ""
        self.all_exact = False
        self.super_id = None

    @property
    def qset(self) -> dict:
        return self._query_set
    
    
    @property
    def geos(self):
        return self._geo_fields
    
    
    @geos.setter
    def geos(self, _geos:list):
        """ Set all of the Geolocational fields """
        self._geo_fields = set(_geos)
        
    @property
    def general(self):
        return self._general
    
    @general.setter
    def general(self, _general):
        self._general = _general

    def exact(self, field, num):
        placeholder = {
            "filter": "number",
            # "operation": "exact",
            "value": {
                "upper": num,
                "lower": num
            }
        }
        self._number_fields.add(field)
        self.qset[field] = placeholder
        return self

    def greater(self, field:str, num):
        placeholder = {
            "filter": "number",
            # "operation": "greater",
            "value": {
                "upper": "+inf",
                "lower": num
            }
        }
        self._number_fields.add(field)
        self.qset[field] = placeholder
        return self
    
    
    def less(self, field, num):
        placeholder = {
            "filter": "number",
            # "operation": "less",
            "value": {
                "upper": num,
                "lower": "-inf"
            }
        }
        self._number_fields.add(field)
        self.qset[field] = placeholder
        return self
    
    
    def between(self, field:str, upper:float, lower:float):
        """
            Search a number field for everything between the two points provided
        """

        placeholder = {
            "filter": "number",
            # "operation": "between",
            "value": {
                "upper": upper,
                "lower": lower
            }
        }
        self._number_fields.add(field)
        self.qset[field] = placeholder
        return self
    
    
    def near(self, field:str, _long:float, _lat:float, distance:float=1, metric:str="km"):
        """ Get all of the geo fields and generate a query where everything in proximity will be picked up. """
        placeholder = {
            "filter": "geo",
            "value": {
                "long": _long,
                "lat": _lat,
                "distance": distance,
                "metric": metric
            }
        }
        self._geo_fields.add(field)
        self.qset[field] = placeholder
        return self
    
    
    def boolean(self, field:str, is_true=False):
        placeholder = {
            "filter": "boolean"
        }
        if is_true:
            placeholder['value'] = "TRUE"
        else:
            placeholder['value'] = "FALSE"
        self._boolean_fields.add(field)
        self.qset[field] = placeholder
        return self
    
    def tags(self, field, tags:list, op="OR"):
        placeholder = {
            "filter": "tags",
            "operation": op,
            "value": {
                "tags": tags
            }
        }
        self._tag_fields.add(field)
        self.qset[field] = placeholder
        return self
    


    def term(self, field, _term:str, is_exact=False):
        updated_term = re.sub('[^a-zA-Z0-9\n\.|\*|\@|\|\_]', ' ', _term)
        placeholder = {
            "filter": "text",
            "is_exact": is_exact,
            "value":  updated_term
        }
        self._text_fields.add(field)
        self.qset[field] = placeholder
    
    def insert_by_type_str(self, _type_str:str, field:str, value):
        if _type_str == "BOOL":
            self.boolean(field, is_true=value)
        elif _type_str == "NUMERIC":
            self.exact(field, value)
        elif _type_str == "TEXT":
            self.term(field, value)
        elif _type_str == "TAG":
            self.tags(field, value)


    def _single_text(self, field:str) -> str:
        """ """

        current = self.qset[field]
        is_exact = current.get('is_exact', False)
        _term = current.get('value', "")
        
        if is_exact or self.all_exact:
            return f'@{field}:\"{_term}\"'
        return f"@{field}:{_term}"

    def _process_text_filter(self) -> str:
        text_lists = [self._single_text(field) for field in self._text_fields]
        joined = " ".join(text_lists)
        joined = joined.strip()
        return joined


    def _single_bool(self, field:str) -> str:
        """ """

        current = self.qset[field]
        _term = current.get('value', "")
        return f"@{field}:\"{_term}\""

    def _process_boolean(self) -> str:
        """ Do an exact match on all boolean values """
        bool_list = [self._single_bool(field) for field in self._boolean_fields]
        joined = " ".join(bool_list)
        return joined
    
    def _single_geo(self, field:str):
        """ Turn a geospacial query into a string """
        current = self.qset[field].get("value")

        lon = current.get("long")
        lat = current.get("lat")
        dist = current.get("distance")
        metric = current.get("metric")
        _geo = f"@{field}:[{lon} {lat} {dist} {metric}]"
        return _geo

    def _process_geo_filter(self) -> str:
        geo_list = [self._single_geo(field)  for field in self._geo_fields]
        joined = " ".join(geo_list)
        return joined
    
    def _single_tag(self, field:str):
        current = self.qset[field]
        values = current.get("value")
        _tags = values.get("tags")
        if len(_tags) == 0:
            return ""
        join_val = "|"
        _joined = join_val.join(_tags)
        _joined = _joined.strip()
        _tag_filter = f"@{field}: {{ {_joined} }}"
        return _tag_filter

    def _process_tag_filter(self) -> str:
        tag_list = [self._single_tag(field) for field in self._tag_fields]
        joined = " ".join(tag_list)
        trimmed = joined.strip()
        return trimmed
    

    def _single_num(self, field):
        current = self.qset[field]

        _value = current.get("value")
        up = _value.get("upper")
        down = _value.get("lower")
        if up == down:
            return f"@{field}:[{up} {up}]"
        
        return f"@{field}: [{down} {up}]"
        


    def _process_number_filter(self):
        tag_list = [self._single_num(field) for field in self._number_fields]
        joined = " ".join(tag_list)
        return joined
    
    """ 
        Convert dictionaries into something we can digest
    """

    def convert_geo_dict(self, name:str, _dict:dict):
        """ Convert a geospecific dictionary """
        values = _dict['values']
        if not is_valid_geo(values):
            logger.error(values)
            return
        _long = values.get("long")
        _lat = values.get("lat")
        _distance = values.get("distance")
        _metric = values.get("metric", "km")
        self.near(field=name, _long=_long, _lat=_lat, distance=_distance, metric=_metric)
    
    def convert_boolean_dict(self, name, _dict):
        pass

    def convert_tags_dict(self, name, _dict):
        pass
    
    def convert_text_dict(self, name:str, _dict:dict):
        """
            1. Check to see if the  dictionary has all of the required components
        """
        values = _dict['values']
        if not is_valid_text(values):
            logger.error(values)
            return
        _term = values.get("term")
        _is_exact = values.get("is_exact", False)
        self.term(field=name, _term=_term, is_exact=_is_exact)

    def convert_numeric_dict(self, name:str, _dict:dict):
        """
            1. Check to see which kind of query we're doing: greater, lesser, between
            2. Check to see if we have dictionary values required to form a query based on each
            3. If it fits, we'd create a query using the above functions 
        """
        values = _dict['values']
        if not is_valid_numeric(values):
            logger.error(values)
            return
        _operation = values.get("operation")
        _up = values.get("upper", 0)
        _low = values.get("lower", 0)

        if _operation == "greater":
            self.greater(name, _low)
        elif _operation == "lesser":
            self.less(name, _up)
        elif _operation == "between":
            self.between(name, _up, _low)
        elif _operation == "exact":
            self.exact(name, _low)


    def from_dict(self, name:str, item:dict):
        """ Take a queryable dictionary and find a query for it"""
        # If this isn't a filter, we're not going to add it into a query
        if not item.get("is_filter"):
            logger.debug("We're not creating a search query from this record")
            return
        
        _type = str(item.get("type")).upper()
        
        if _type == "GEO":
            self.convert_geo_dict(name, item)
        elif _type == "BOOL":
            self.convert_boolean_dict(name, item)
        elif _type == "TEXT":
            self.convert_text_dict(name, item)
        elif _type == "TAGS":
            self.convert_tags_dict(name, item)
        else:
            self.convert_numeric_dict(name, item)


    def build(self):
        """Builds a query to be executed"""
        processed_text = self._process_text_filter()
        processed_bool = self._process_boolean()
        processed_num = self._process_number_filter()
        processed_tags = self._process_tag_filter()
        processed_geo = self._process_geo_filter()
        joined_query_string = " ".join([processed_text, processed_tags, processed_num, processed_bool, processed_geo])
        final_joined = self.general + joined_query_string
        return final_joined

    def build_exact(self):
        self.all_exact = True
        processed = self.build()
        self.all_exact = False
        return processed.strip()

class InsertBuilder(object):
    """
        # InsertBuilder

        Builds a redis query given a set of builder parameters
    """
    def __init__(self):
        self._insert_dict = {}
        self.doc_id = ""
        self.is_replacement = False
        self.is_sub = False
        self.super_id = None


    def add_field(self, name, field_type, **values):
        """ Adds a field with the values inside of the kwargs field. Will go through specific formatting."""
        return self


    
    def boolean(self, field:str, is_true=False):
        new_field = inserter.boolean_process(field, is_true=is_true)
        self._insert_dict.update(new_field)
        return self
    
    def tags(self, field, tags:list, op="OR"):
        new_field = inserter.list_process(field, tags)
        self._insert_dict.update(new_field)
        return self
    
    def within(self, field, _within:list):
        """ 
            # WITHIN
            Essentially a `WHERE x IN` command for SQL

            WHERE x IN ('foo', 'bar','hello world')
        """
        pass

    def term(self, field, _term:str, is_exact=False):
        new_field = inserter.text_process(field, _term, is_exact)
        self._insert_dict.update(new_field)
        return self

    

    def exact(self, field:str, num):
        """ Set's an exact number """
        self._insert_dict[field] = num
        return self
  


    def convert_geo_dict(self, name:str, item:dict):
        new_item = inserter.geo_process_dict(name, item)
        self._insert_dict.update(new_item)

    def convert_text_dict(self, name:str, item:dict):
        new_item = inserter.text_process_dict(name, item)
        self._insert_dict.update(new_item)
    

    def convert_numeric_dict(self, name:str, item:dict):
        new_item = inserter.num_process_dict(name, item)
        self._insert_dict.update(new_item)


    def from_dict(self, name:str, item:dict):
        """ Get the values from the dictionary for something that's insertable. """
        if not item.get("is_filter"):
            logger.debug("We're not creating a search query from this record")
            return
        
        # logger.debug(item)
        _type = str(item.get("type")).upper()
        
        if _type == "GEO":
            self.convert_geo_dict(name, item)
        elif _type == "TEXT":
            self.convert_text_dict(name, item)
        else:
            self.convert_numeric_dict(name, item)

    def insert_by_type_str(self, _type_str:str, field:str, value):
        if _type_str == "BOOL":
            self.boolean(field, is_true=value)
        elif _type_str == "NUMERIC":
            self.exact(field, value)
        elif _type_str == "TEXT":
            self.term(field, value)
        elif _type_str == "TAG":
            self.tags(field, value)



    def build(self):
        """ After all of the dictionaries are set we create a doc_id (for the given index) and we create one"""
        # if 
        self.doc_id = uuid.uuid4().hex
        if self.super_id is not None:
            self._insert_dict['super_id'] = self.super_id
        return self._insert_dict

    def reset(self):
        """ Reset all of the items that are inside of object."""
        self._insert_dict = {}
        self.super_id = None
